#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import unittest

import numpy as np

from numpy.testing import assert_raises

from ..support import array_assert


class MITests(unittest.TestCase):
    def test_mutual_information_bad(self):
        """Test that we raise an exception with bad MI input"""
        from ..stats import mutual_information
        assert_raises(ValueError, mutual_information,
                      np.zeros(1000,), np.zeros(1000, ))

    def test_mutual_information(self):
        """Test that we calculate a test MI value correctly (log2)"""
        from ..stats import mutual_information

        x1 = np.sin(2 * np.pi * np.linspace(0, 2, 1000))[..., np.newaxis].T
        x2 = np.cos(2 * np.pi * np.linspace(0, 2, 1000))[..., np.newaxis].T

        ret = mutual_information(x1, x2, r=(-1.0, 1.0), nbins=10)

        array_assert(ret, np.array([1.50009220434754353413]))

    def test_mutual_information_calc_range(self):
        """Test that we calculate MI value correctly when using auto range"""
        from ..stats import mutual_information

        x1 = np.sin(2 * np.pi * np.linspace(0, 2, 1000))[..., np.newaxis].T
        x2 = np.cos(2 * np.pi * np.linspace(0, 2, 1000))[..., np.newaxis].T

        ret = mutual_information(x1, x2, nbins=10)

        array_assert(ret, np.array([1.50011373288843841145]), decimal=5)


class DWTests(unittest.TestCase):
    def test_durbin_watson_1d(self):
        """Test Durbin-Watson calculation from 1D data"""
        from ..stats import durbin_watson

        dat = np.arange(1, 101)
        ret = durbin_watson(dat)
        correct = np.array([2.9259642382148664261e-04])
        array_assert(ret.real, correct)

    def test_durbin_watson_2d(self):
        """Test Durbin-Watson calculation from 2D data"""
        from ..stats import durbin_watson

        dat = np.arange(1, 101)[np.newaxis, slice(None)]
        ret = durbin_watson(dat)
        correct = np.array([2.9259642382148664261e-04])

        array_assert(ret.real, correct)

    def test_durbin_watson_3d(self):
        """Test Durbin-Watson calculation from 3D data"""
        from ..stats import durbin_watson

        dat = np.zeros((1, 100, 2))
        dat[0, :, 0] = np.arange(1, 101)
        dat[0, :, 1] = np.arange(101, 201)
        ret = durbin_watson(dat)
        correct = np.array([2.9259642382148664261e-04,
                            4.2157259352311195991e-05])

        array_assert(ret.real, correct)

    def test_durbin_watson_nd(self):
        """Test Durbin-Watson calculation fails with > 3D data"""
        from ..stats import durbin_watson

        dat = np.zeros((1, 100, 1, 1))
        assert_raises(ValueError, durbin_watson, dat)


class ResidualsTests(unittest.TestCase):
    def test_rsquared_from_residuals_bad_input(self):
        """Test rsquared_from_residuals calculation with bad input"""
        from ..stats import rsquared_from_residuals

        data = np.zeros((1, 100))
        residuals = np.zeros((2, 100))

        assert_raises(ValueError, rsquared_from_residuals, data, residuals)

    def test_rsquared_from_residuals_int(self):
        """Test rsquared_from_residuals calculation from 1D data (int)"""
        from ..stats import rsquared_from_residuals

        data = np.arange(1, 101)
        residuals = data * 0.5
        correct = 0.75

        # Both per_signal True and False should be the same here
        ret = rsquared_from_residuals(data, residuals, per_signal=True)
        assert(ret == correct)

        ret = rsquared_from_residuals(data, residuals, per_signal=False)
        assert(ret == correct)

    def test_rsquared_from_residuals_1d(self):
        """Test rsquared_from_residuals calculation from 1D data (float)"""
        from ..stats import rsquared_from_residuals

        data = np.arange(1, 101).astype(np.float64)
        residuals = data * 0.5
        correct = 0.75

        # Both per_signal True and False should be the same here
        ret = rsquared_from_residuals(data, residuals, per_signal=True)
        assert(ret == correct)

        ret = rsquared_from_residuals(data, residuals, per_signal=False)
        assert(ret == correct)

    def test_rsquared_from_residuals_2d(self):
        """Test rsquared_from_residuals calculation from 2D data"""
        from ..stats import rsquared_from_residuals

        data = np.zeros((2, 100))
        data[0, :] = np.arange(1, 101)
        data[1, :] = np.arange(1, 101)
        residuals = data.copy()
        residuals[0, :] *= 0.5
        residuals[1, :] *= 0.75

        # Check both combined and per-signal calculation
        correct_per = np.array([0.75, 0.4375])
        correct_comb = 0.59375
        ret = rsquared_from_residuals(data, residuals, per_signal=True)
        array_assert(ret, correct_per)

        ret = rsquared_from_residuals(data, residuals, per_signal=False)
        assert(ret == correct_comb)

    def test_rsquared_from_residuals_3d(self):
        """Test Durbin-Watson calculation from 3D data"""
        from ..stats import rsquared_from_residuals

        data = np.zeros((2, 100, 2))
        data[0, :, :] = np.arange(1, 101)[..., np.newaxis]
        data[1, :, :] = np.arange(1, 101)[..., np.newaxis]
        residuals = data.copy()
        residuals[0, :, :] *= 0.5
        residuals[1, :, :] *= 0.75

        # Check both combined and per-signal calculation
        correct_per = np.array([0.75, 0.4375])
        correct_comb = 0.59375
        ret = rsquared_from_residuals(data, residuals, per_signal=True)
        array_assert(ret, correct_per)

        ret = rsquared_from_residuals(data, residuals, per_signal=False)
        assert(ret == correct_comb)

    def test_rsquared_from_residuals_nd(self):
        """Test Durbin-Watson calculation fails with > 3D data"""
        from ..stats import rsquared_from_residuals

        data = np.zeros((1, 100, 1, 1))
        assert_raises(ValueError, rsquared_from_residuals, data, data)


class MahalnobisTests(unittest.TestCase):
    def test_mahalanobis_origin(self):
        """Test mahalanobis calculation to the origin"""
        from ..stats import mahalanobis

        # Set up a co-ordinate system
        dat = np.array([np.arange(1, 101),
                        np.sin(np.arange(1, 101)),
                        np.tan(np.arange(1, 301, 3))])

        # Pre-compute the covariance matrix
        sigma = np.cov(dat)

        # Feed in a couple of points of known distance
        pts = np.array([[1.0, 1, 1], [2, 2, 2]]).T

        correct = np.array([2.0136770397279, 8.0547081589117])

        ret = mahalanobis(pts, sigma=sigma)

        # An inverse is involved so allow for some difference
        array_assert(ret, correct, decimal=10)

    def test_mahalanobis_point(self):
        """Test mahalanobis calculation to another point"""
        from ..stats import mahalanobis

        # Set up a co-ordinate system
        dat = np.array([np.arange(1, 101),
                        np.sin(np.arange(1, 101)),
                        np.tan(np.arange(1, 301, 3))])

        # Pre-compute the covariance matrix
        sigma = np.cov(dat)

        # Feed in a couple of points of known distance
        pts_1 = np.array([[1.0, 1, 1], [2, 2, 2]]).T
        pts_2 = np.array([[1.0, 1, 1], [0, 0, 0]]).T

        correct = np.array([0.0, 8.0547081589117])

        ret = mahalanobis(pts_1, pts_2, sigma=sigma)

        # An inverse is involved so allow for some difference
        array_assert(ret, correct, decimal=8)

    def test_mahalanobis_point_error(self):
        """Test mahalanobis calculation to another point"""
        from ..stats import mahalanobis

        # Feed in a couple of points of known distance
        pts_1 = np.array([[1.0, 1, 1], [2, 2, 2]]).T
        pts_2 = np.array([[1.0, 1, 1], [0, 0, 0], [3, 3, 3]]).T

        assert_raises(ValueError, mahalanobis, pts_1, pts_2)

    def test_mahalanobis_compute_sigma(self):
        """Test mahalanobis calculation when computing sigma"""
        from ..stats import mahalanobis

        # Set up some co-ordinates
        dat = np.array([np.arange(1, 5),
                        np.sin(np.arange(1, 5)),
                        np.tan(np.arange(1, 12, 3))])

        ret = mahalanobis(dat)

        correct = np.array([46738.68983456, 47169.93175778,
                            46209.49548723, 47068.83525808])

        # An inverse is involved so allow for some difference
        array_assert(ret, correct, decimal=5)


class ProfileLikelihoodTests(unittest.TestCase):
    def test_profile_likelihood(self):
        """Test profile likelihood function"""
        from ..stats import profile_likelihood
        # TODO: This could do with a better external validation

        dat = np.zeros((10, 1))
        dat[0:5, 0] = np.arange(5, 0, -1)
        dat[5:,  0] = np.arange(0.8, -0.1, -0.2)

        correct = np.array([0.00000000000000000000,
                            0.00000000000000000000,
                            3.21859683762712212030,
                            5.06387510343866775742,
                            4.36930553905676788418,
                            2.50420226265147682909,
                            1.67250106964246647578,
                            1.15155564748249972240,
                            0.76855251039401584201,
                            0.00000000000000000000])

        ret = profile_likelihood(dat)

        array_assert(ret, correct, decimal=10)

    def test_profile_likelihood_bad_input(self):
        """Test profile likelihood function with bad input"""
        from ..stats import profile_likelihood

        dat = np.zeros((10, 1, 2))

        assert_raises(ValueError, profile_likelihood, dat)

    def test_find_cov(self):
        """Test find_cov"""
        from ..stats import find_cov

        # Datasets should be signals, samples)
        d1 = np.array([[0, 4, 6, 8],
                       [-1, 1, 3, 5]])

        d2 = np.array([[1, -1, 1, -1],
                       [3, -3, 3, -3]])

        expected = np.array([[-1.5+0j, -4.5+0j],
                             [-1.0+0j, -3.0+0j]])

        res = find_cov(d1, d2, ddof=0)

        array_assert(res, expected)

        # Routine should demean the data across samples, so add an offset
        # and compare

        d1 += 100

        res = find_cov(d1, d2, ddof=0)

        array_assert(res, expected)

        # DDOF of None should multiply by 4 (number of time points)
        res = find_cov(d1, d2, ddof=None)

        array_assert(res, expected * 4.0)

    def test_find_cov_multitrial(self):
        """Test find_cov_multitrial"""
        from ..stats import find_cov_multitrial

        # Datasets should be signals, samples, trials)
        d1 = np.array([[[0, 2, 4, 6], [-1, 1, 3, 5]],
                       [[0, 2, 4, 8], [8, -8, 8, -8]]])

        d2 = np.array([[[0, 1, 2, 3], [-1, 0, 1, 2]],
                       [[9, -9, 9, -9], [11, -11, 11, -11]]])

        expected = np.array([[0.50+0j, 0.0+0j], [1.75+0j, 9.50+0j]])

        res = find_cov_multitrial(d1, d2)

        array_assert(res, expected)

        # Routine should demean the data across samples, so add an offset
        # and compare

        d1 += 100

        res = find_cov_multitrial(d1, d2)

        array_assert(res, expected)

        # DDOF of None should be the same as 1 (default) in this case
        res = find_cov_multitrial(d1, d2, ddof=None)

        array_assert(res, expected)

        # Changing the DDOF to 0 should just scale by 2 in this case
        res = find_cov_multitrial(d1, d2, ddof=0)

        array_assert(res, expected / 2.0)

    def test_stability_index(self):
        """Test the stability index routine"""
        from ..stats import stability_index

        # The SI of an A matrix which is identity at lag 0
        # and zero everywhere else should be 0

        A1 = np.eye(2)
        A2 = np.zeros((2, 2))
        A = np.stack([A1, A2, A2], 2)

        res = stability_index(A)

        assert(res == 0.0)

        # The SI of an A matrix which is identity at lag 1
        # and zero everywhere else should be 1.0

        A = np.stack([A2, A1, A2], 2)

        res = stability_index(A)

        assert(res == 1.0)



from scipy import signal
import numpy as np


def morlet(x, freqs, sample_rate, window_len=4, ncycles=5, ret_basis=False,
           ret_mode='power', normalise=False):
    """Compute a morlet wavelet time-frequency transform on a univariate dataset.

    Parameters
    ----------
    x : vector array_like
        Time-series to compute wavelet transform from.
    freqs : array_like
        Array of frequency values in Hz
    sample_rate : scalar
        Sampling frequency of data in Hz
    window_len : scalar
        Length of wavelet window
    ncycles : int
        Width of wavelets in number of cycles
    ret_basis : bool
        Boolean flag indicating whether to return the basis set alongside the transform.
    ret_mode : {'power','amplitude','complex'}
        Flag indicating whether which form of the wavelet transform to return.

    Returns
    -------
    2D array
        Array containing morlet wavelet transformed data [nfreqs x nsamples]
    """

    cwt = np.zeros((len(freqs), *x.shape[:]), dtype=complex)

    # Get morlet basis
    mlt = get_morlet_basis(freqs, ncycles, sample_rate, normalise=normalise)

    for ii in range(len(freqs)):
        a = signal.convolve(x, mlt[ii].real, mode='same', method='fft')
        b = signal.convolve(x, mlt[ii].imag, mode='same', method='fft')
        cwt[ii, ...] = a+1j*b

    if ret_mode == 'power':
        cwt = np.power(np.abs(cwt), 2)
    elif ret_mode == 'amplitude':
        cwt = np.abs(cwt)
    elif ret_mode != 'complex':
        raise ValueError("'ret_mode not recognised, please use one of {'power','amplitude','complex'}")

    if ret_basis:
        return cwt, mlt
    else:
        return cwt


def get_morlet_basis(freq, ncycles, sample_rate, normalise=False, win_len=5):
    """Compute a morlet wavelet basis set based on specified parameters.

    Parameters
    ----------
    freq : array_like
        Array of frequency values in Hz
    ncycles : int
        Width of wavelets in number of cycles
    sample_rate : scalar
        Sampling frequency of data in Hz
    normalise : {None,'simple','tallon','wikipedia','mne'}
        Flag indicating which normalisation factor to apply to the wavelet basis.
    win_len : float
        Window length duration factor

    Returns
    -------
    list of vector arrays
        Complex valued arrays containing morlet wavelets

    """

    m = []
    for ii in range(len(freq)):
        # Sigma controls the width of the gaussians applied to each wavelet. This
        # is adaptive for each frequency to match ncycles
        sigma = ncycles / (2*np.pi*freq[ii])

        # Compute time vector for this wavelet
        t = np.arange(-win_len*sigma, win_len*sigma, 1/sample_rate)

        # Compute oscillatory component
        wave = np.exp(2*np.pi*1j*t*freq[ii])

        # Compute gaussian-window component
        gauss = np.exp((-(t/2)**2) / (2*sigma**2))

        # Make wavelet
        mlt = wave * gauss

        if normalise == 'simple':
            # Set simple normalisation (output amplitude should match
            # oscillation amplitude)
            mlt = 2 * mlt / np.abs(mlt).sum()
        elif normalise == 'tallon':
            # Set normalisation factor from Tallon-Baudry 1997
            A = (sigma*np.sqrt(np.pi))**(-1/2)
            mlt = A * mlt
        elif normalise == 'wikipedia':
            # Set normlisation from wikipedia: https://en.wikipedia.org/wiki/Morlet_wavelet
            A = (1 + np.exp(-sigma**2) - 2 * np.exp(-3/4 * sigma**2)) ** -0.5
            A = np.pi**(-.25) * A
            mlt = A * mlt
        elif normalise == 'mne':
            # Set normalisation step from MNE-python
            # https://github.com/mne-tools/mne-python/blob/master/mne/time_frequency/tfr.py#L98
            mlt = mlt / (np.sqrt(0.5) * np.linalg.norm(mlt.ravel()))

        m.append(mlt)
    return m



#!/usr/bin/python

import numpy as np
from scipy import linalg


def closest_orthogonal(A, verbose=False):
    """Implementation of the closest-orthonormal orthogonalisation method
    defined in [Colclough2015]_

    Parameters
    ----------
    A : ndarray
        Data matrix to orthogonalise, of size [channels, samples]
    verbose : bool
        Flag indicating whether to show detailed output (Default value = False)

    Returns
    -------

    data : numpy.ndarray [channels, samples] of orthogonalised data

    d : numpy.ndarray

    rho : numpy.ndarray

    W : numpy.ndarray

    """

    A = A.T

    U, S, V = linalg.svd(A, full_matrices=False)
    tol = S.max() * max(A.shape) * np.finfo(S.dtype).eps

    # Settings
    max_iter = 300
    it = 0

    # Initialise d to ones, we start with the closest orthonormal matrix
    d = np.ones((A.shape[1],))

    rho = np.zeros((max_iter, 1)) * np.nan

    while it < max_iter:

        # Find orthonormal polar factor
        V, _ = symmetric_orthonormal(A.dot(np.diag(d)).T)

        # Minimise rho w.r.t d
        d = np.diag(np.conj(A).T.dot(V.T))

        # New best estimre
        L = V.T.dot(np.diag(d))

        # Error term
        E = A - L
        rho[it] = np.sum(np.diag(E.T.dot(E)))

        if verbose:
            print('iter: %4d\trho: %g' % (it, rho[it]))

        if it > 1 and np.abs(rho[it] - rho[it-1]) <= tol:
            break

        it += 1

    # Finally calculate the linear operator
    _, W = symmetric_orthonormal(A.dot(np.diag(d)).T)
    W = np.diag(d).dot(W).dot(np.diag(d))

    if it == max_iter:  # pragma: nocover
        # ? Turn this into an exception?
        print("MaxIterationsHit: the results my not be optimal")

    # Tidy up the rho vector
    rho = rho[~np.isnan(rho)]

    return L.T, d, rho, W


def symmetric_orthonormal(A, maintain_mag=False):
    """Implement the [Colclough2015]_ algorithm for multivariate leakage
    correction in MEG.  Also see the OHBA toolbox
    (https://github.com/OHBA-analysis/MEG-ROI-nets) and
    http://empslocal.ex.ac.uk/people/staff/reverson/uploads/Site/procrustes.pdf

    Parameters
    ----------
    A : ndarray
        Data matrix to orthogonalise, of size [channels, samples]
    verbose : bool
        Flag indicating whether to show detailed output (Default value = False)
    maintain_mag : bool
        Flag indicating whether to maintain magnitudes (Default value = False)

    Returns
    -------

    data : numpy.ndarray
          [channels, samples] of orthogonalised data

    W : numpy.ndarray

    """

    Adecom = A.T

    # Do we need to adjust for magnitude?
    if maintain_mag:
        D = np.diag(np.sqrt(np.diag(A.dot(A.T))))
        Adecom = A.T.dot(D)

    # Compute svd
    U, S, V = linalg.svd(Adecom, full_matrices=False)

    # This is equivalent to np.linalg.matrix_rank
    tol = S.max() * max(A.shape) * np.finfo(S.dtype).eps
    rank = sum(S > tol)
    if rank < A.shape[0]:
        raise Exception('Input is not full rank!')

    # Not conj(transpose(V)) in python as V is already transposed
    Lnorm = U.dot(V.conj())
    W = V.T.dot(np.diag(1.0 / S)).dot(V.conj())

    if maintain_mag:
        Lnorm = Lnorm.dot(D)
        W = D.dot(W).dot(D)

    # Ensure that we return in the same order as we were input
    return Lnorm.T, W


def innovations(A, order, mvar_class=None, verbose=False):
    """Implementation of the innovations-orthogonalisation defined in
    [Pascual-Marqui2017]_ . This creates the mixing matrix on the input data
    after variance explained by an MVAR model is removed. The orthogonalisation
    is defined on the residuals but applied to the raw data.

    Parameters
    ----------
    A : ndarray
        Data matrix to orthogonalise, of size [channels, samples]
    order : int
        Model order used during MVAR model fit
    mvar_class : sails LinearModel class
        Model class to do model fitting (Default value = None)
    verbose : bool
        Flag indicating whether to show detailed output (Default value = False)

    Returns
    -------
    ndarray :
        Orthogonalised data matrix, of size [channels, samples]


    """

    if mvar_class is None:
        from .modelfit import VieiraMorfLinearModel
        mvar_class = VieiraMorfLinearModel()

    # Fit AR model to raw data
    delay_vect = np.arange(order)
    m = mvar_class.fit_model(A[:, :, None], delay_vect)

    # Get innovations time-series (aka residuals)
    from sails.modelfit import get_residuals
    innovs = get_residuals(A[:, :, None], m.parameters, delay_vect)

    # Orthogonalise innovations
    orth_innovs, _, _, _ = closest_orthogonal(innovs[:, :, 0])

    # Channels x Channels mixing matrix
    M = np.linalg.pinv(orth_innovs.T).dot(innovs[:, :, 0].T)

    # Apply (un)mixing matrix and return orthogonalised data
    return A.T.dot(M).T



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import numpy as np

from numpy.linalg import inv, norm, eig

from scipy.signal import fftconvolve

__all__ = []


def find_cov_multitrial(a, b, ddof=1):
    """Estimate the average covariance between two datasets across many trials

    Parameters
    ----------
    a : ndarray
        a multivariate time series of shape nsignals x nsamples x ntrials
    b : ndarray
        a multivariate time series of shape nsignals x nsamples x ntrials
    ddof : int
        if None, no normalisation will be performed.  If an integer,
        covariance will be normalised by (nsamples - ddof) (Default value = 1)

    Returns
    -------
    type
        A covariance matrix between a and b of shape nsignals x nsignals

    """

    if a.dtype is not b.dtype:
        raise TypeError('Data types of input a ({0}) and b ({1}) not matched')

    nsignals = a.shape[0]
    nsamples = a.shape[1]
    ntrials = a.shape[2]

    # Make sure datasets are demeaned before calculating covariance
    a = a - a.mean(axis=1)[slice(None), np.newaxis]
    b = b - b.mean(axis=1)[slice(None), np.newaxis]

    cov = np.zeros((nsignals, nsignals, ntrials), dtype=a.dtype)

    for i in range(ntrials):
        b_h = b[:, :, i].T.conj()
        cov[:, :, i] = a[:, :, i].dot(b_h)

    if ddof is not None:
        return cov.mean(axis=2) / (nsamples - ddof)
    else:
        return cov.mean(axis=2)


__all__.append('find_cov_multitrial')


def find_cov(a, b, ddof=1):
    """Estimate the covariance between two datasets.

    Parameters
    ----------
    a : ndarray
        a multivariate time series of shape nsignals x nsamples
    b : ndarray
        a multivariate time series of shape nsignals x nsamples
    ddof : int
        if None, no normalisation will be performed.  If an integer,
        covariance will be normalised by (nsamples - ddof) (Default value = 1)

    Returns
    -------
    type
        A covariance matrix between a and b of shape nsignals x nsignals

    """

    if a.dtype is not b.dtype:
        raise TypeError('Data types of input a ({0}) and b ({1}) not matched')

    nsamples = float(a.shape[1])

    # Make sure datasets are demeaned before calculating covariance
    a = a - a.mean(axis=1)[slice(None), np.newaxis]
    b = b - b.mean(axis=1)[slice(None), np.newaxis]

    b_h = b.T.conj()
    cov = a.astype(a.dtype).dot(b_h)

    if ddof is not None:
        return cov / (nsamples - ddof)
    else:
        return cov


__all__.append('find_cov')


def mutual_information(A, B, r=None, nbins=21, base=np.log2):
    """Calculate the mutual information between two signals.

    Parameters
    ----------
    A : ndarray
        First signal, of shape [channels x samples]
    B : ndarray
        Second signal, of shape [channels x samples]
    r : tuple (min,max)
        Range of the random variables used for bin calculations.
        If not set, takes the minimum and maximum values from
        the combination of A and B (Default value = None)
    nbins : int
        Number of bins to compute the frequencies within (Default value = 21)
    base : function
        Base to use for MI calculation - defaults to np.log2

    Returns
    -------
    type
        the mutual information within each signal passed

    """

    if A.ndim != 2 or B.ndim != 2:
        raise ValueError("Require 2D arrays for A and B")

    nsignals = A.shape[0]

    if r is None:
        r = (min(A.min(), B.min()), max(A.max(), B.max()))

    MI = np.zeros((nsignals,))

    for s in range(nsignals):
        PA = np.histogram(A[s, :], range=r,
                          bins=nbins)[0].astype(np.float64)
        PB = np.histogram(B[s, :], range=r,
                          bins=nbins)[0].astype(np.float64)
        PAB = np.histogram2d(A[s, :], B[s, :], range=(r, r),
                             bins=nbins)[0].astype(np.float64)

        # Outer product
        Pmarg = np.dot(np.atleast_2d(PA).T, np.atleast_2d(PB))

        # Scale the PAB and Pmarg matrices
        PAB /= PAB.sum()
        Pmarg /= Pmarg.sum()

        # Handle 0s properly
        idx_x, idx_y = np.nonzero(PAB / Pmarg)

        tmp = base(PAB[idx_x, idx_y] / Pmarg[idx_x, idx_y])
        MI[s] = np.sum(PAB[idx_x, idx_y] * tmp)

    return MI


__all__.append('mutual_information')


def durbin_watson(residuals, step=1):
    """Test for autocorrelation in the residuals of a model (see [Durbin1950]_
    and [Durbin1951]_)

    The result is a value between 0 and 4. A value of 0 indicates a positive
    autocorrelaiton, 2 indicates no correlation and a value of 4 indicates a
    negative autocorrelation.

    If many trials are found the statistic is estimated for each trial
    separately.

    Parameters
    ----------
    residuals : ndarray
        Residuals of shape [channels, samples, ntrials]
        If only two dimensions, will be treated as
        [channels, samples]
        If only one dimension, will be treated as [samples]
    step : int
        step-size to use through residuals when calculating DW.
        Defaults to 1.

    Returns
    -------
    type
        durbin-watson index

    """

    if residuals.ndim == 1:
        residuals = residuals[np.newaxis, slice(None), np.newaxis]
    elif residuals.ndim == 2:
        residuals = residuals[..., np.newaxis]
    elif residuals.ndim != 3:
        raise ValueError("residuals must be 1, 2 or 3D")

    # Calculation needs to be performed in the complex domain
    residuals = residuals.astype(np.complex)
    d_resid = residuals[:, :-step, :] - residuals[:, step:, :]

    # Combine all trials by flattening axes 0 and 1
    d_resid = d_resid.reshape(d_resid.shape[0]*d_resid.shape[1],
                              d_resid.shape[2])

    residuals = residuals.reshape(residuals.shape[0]*residuals.shape[1],
                                  residuals.shape[2])

    num = d_resid.T.dot(d_resid).real
    den = (residuals.T.dot(residuals)).real

    return np.diag(num / den).astype(np.float64)


__all__.append('durbin_watson')


def percent_consistency(data, residuals):
    """Estimate how well a model retains the correlational structure of a dataset.

    This uses the percent consistency measure outlined in [Ding2000]_. All
    auto and cross correlations within a multivariate data set are computed and
    compared to the auto and cross correlations of the data as predicted by a
    model (in this case the data - model residuals, as these are easily
    obtained).

    Parameters
    ----------
    data : ndarray
        An array containing the observed data, of shape [nsignals x
        nsamples x ntrials]
    residuals : ndarray
        An array containing the remaining variance of data after
        model fitting, of shape [nsignals x nsamples x ntrials]

    Returns
    -------
    type
        A value between 0 and 100 indicating the percentage of the
        autocorrelation in the data that is retained in the model.

    """

    if data.shape != residuals.shape:
        raise ValueError("data and residuals must be the same shape")

    if data.ndim == 1:
        data = data[np.newaxis, ..., np.newaxis]
        residuals = residuals[np.newaxis, ..., np.newaxis]
    elif data.ndim == 2:
        data = data[..., np.newaxis]
        residuals = residuals[..., np.newaxis]
    elif residuals.ndim != 3:
        raise ValueError("data and residuals must be 1, 2 or 3D")

    resid = residuals

    # Preallocate arrays for the correlations

    # Real
    R_r = np.zeros((data.shape[0], data.shape[0],
                    data.shape[2], data.shape[1]), dtype=np.complex)
    # Prediction
    R_p = np.zeros((data.shape[0], data.shape[0],
                    data.shape[2], data.shape[1]), dtype=np.complex)

    # Compute auto and cross correlations
    for i in range(data.shape[0]):  # node
        for j in range(data.shape[0]):  # node
            for t in range(data.shape[2]):  # epoch
                # array flipped convolution which is faster for large arrays.
                # This is an fft based method for computing the cross
                # correlation
                R_r[i, j, t, :] = fftconvolve(data[i, :, t],
                                              data[j, ::-1, t],
                                              mode='same')
                R_p[i, j, t, :] = fftconvolve(data[i, :, t]-resid[i, :, t],
                                              data[j, ::-1, t]-resid[j, :, t],
                                              mode='same')

    # Reshape into a vector
    R_r = R_r.reshape(-1)
    R_p = R_p.reshape(-1)

    # Calculate and return PC as defined in Ding et al 2000 - eq 12
    num = norm(R_p - R_r)
    denom = norm(R_r)

    percent_consistency = (1 - (num / denom)) * 100

    return np.float64(percent_consistency)


__all__.append('percent_consistency')


def rsquared_from_residuals(data, residuals, per_signal=False):
    """Estimate the variance explained by a model from the original data and the
    model residuals

    If data and residuals are 2D, they will be treated as [channels, nsamples]

    If data and residuals are 1D, they will be treated as [nsamples]

    Parameters
    ----------
    data :  ndarray
        Array containing the original data,
        of shape [channels, nsamples, ntrials]
    residuals : ndarray
        Array containing the model residuals,
        of shape [channels, nsamples, ntrials]
    per_signal : bool
        Boolean indicating whether the variance explained
        should be computed separately per channel (Default value = False)

    Returns
    -------
    type
        Value between 0 and 1 indicating the proportion of variance
        explained by the model.

    """

    if data.shape != residuals.shape:
        raise ValueError("data and residuals must be the same shape")

    if data.ndim == 1:
        data = data[np.newaxis, ..., np.newaxis]
        residuals = residuals[np.newaxis, ..., np.newaxis]
    elif data.ndim == 2:
        data = data[..., np.newaxis]
        residuals = residuals[..., np.newaxis]
    elif residuals.ndim != 3:
        raise ValueError("data and residuals must be 1, 2 or 3D")

    if per_signal:
        dat = (data**2.0).sum(axis=1).sum(axis=1)
        res = (residuals**2.0).sum(axis=1).sum(axis=1)
    else:
        dat = (data**2.0).sum()
        res = (residuals**2.0).sum()

    return np.float64(1 - (res.real / dat.real))


__all__.append('rsquared_from_residuals')


def mahalanobis(X, Y=None, sigma=None):
    """Estimate the Mahalanobis distance of each point in an array across the
    samples dimension.

    Parameters
    ----------
    X : ndarray
        Array containing data of points: shape [channels x samples]
    Y : ndarray (Optional)
        Optional Array containing data of second points: shape
        [channels x samples].  If not provided, the origin will be
        used as the point for comparison. (Default value = None)
    sigma : ndarray (Optional)
        A precomputed covariance matrix to use in the calculation,
        of shape [channels x channels].  If not provided, sigma
        will be calculated from X (Default value = None)

    Returns
    -------
    type
        The Mahalanobis distance for each sample, of shape [samples]

    """

    if Y is not None:
        if Y.shape != X.shape:
            raise ValueError("If Y is provided, must be the same shape as X")

        # Offset the points
        X = X - Y

    # Calculate our covariance matrix if necessary
    if sigma is None:
        sigma = np.cov(X)

    sigma_i = inv(sigma)

    return np.diag(X.T.dot(sigma_i).dot(X))


__all__.append('mahalanobis')


def profile_likelihood(X):
    """Find the 'elbow' point in a curve.

    This function uses the automatic dimensionality selection method from [Zhu2006]_
    to estimate the elbow or inflection point in a given curve.

    Parameters
    ----------
    X : ndarray
        1d array containing the curve to be analysed, of shape [samples]

    Returns
    -------
    type
        The log likelihood that each point on the curve separates the
        distributions to the left and right, of shape [samples]

    """

    # Handle 2D arrays with a singleton second dimension
    X = X.squeeze()
    if X.ndim != 1:
        raise ValueError("Only 1D matrices supported by profile_likelihood")

    llh = np.zeros_like(X)

    for i in range(2, X.shape[0]-1):
        left = X[0: i]
        right = X[i:-1]

        # Common scale parameter
        csp = ((i-1)*left.var() + (len(X)-i-1)*right.var()) / (len(X)-2.)

        left_power = (left - left.mean())**2 / 2*csp
        left_side = np.sum(1.0 / np.sqrt(2*np.pi*csp) * np.exp(-left_power))

        right_power = (right - right.mean())**2 / 2*csp
        right_side = np.sum(1.0 / np.sqrt(2*np.pi*csp) * np.exp(-right_power))

        llh[i] = left_side + right_side

    return llh


__all__.append('profile_likelihood')


def stability_index(A):
    """Compute the stability index as defined in [Lutkephol2006]_ pages 15-16.

    This is a proxy for the stationarity assumption of MVAR models, if the
    magnitude of the largest principal component of the model parameters is
    less than 1, the model is stable and thus, stationary.

    Parameters
    ----------
    A : ndarray
        The MVAR parameter matrix of  size [nchannels x nchannels x model_order]

    Returns
    -------
    type
        The stability index

    """

    nsignals = A.shape[0]
    order = A.shape[2]

    # Form companion matrix
    companion = np.eye(nsignals*(order-1), k=nsignals).T
    companion[:nsignals, :] = A[:, :, 1:].reshape((nsignals,
                                                   nsignals*(order-1)),
                                                  order='F')

    # Find eigenvalues
    evals, evecs = eig(companion)

    # Find the largest magnitude
    stability_index = np.abs(evals).max()

    return np.float64(stability_index)


__all__.append('stability_index')


def stability_ratio(A):
    """Compute the stability ratio as a measure of stationarity

    A stronger test for stationarity than the stability index. Computes the
    ratio between the largest eigenvalue of the parameters and all the others.
    If the largest parameter is larger than all other parameters (indicated by a
    value < 1) we have a super-stable system.

    Parameters
    ----------
    A : ndarray
        The MVAR parameter matrix of  size [nchannels x nchannels x model_order]


    Returns
    -------
    type
        The stability ratio

    """

    nsignals = A.shape[0]
    order = A.shape[2]

    # Form companion matrix
    companion = np.eye(nsignals*(order-1), k=nsignals).T
    companion[:nsignals, :] = A[:, :, 1:].reshape((nsignals,
                                                   nsignals*(order-1)),
                                                  order='F')

    # Find eigenvalues
    evals, evecs = eig(companion)

    # Take the magnitude
    evals = np.abs(evals)

    # Get the smaller eigen values
    evals_smaller = evals[np.where(evals != evals.max())]

    # Express the ratio
    stability_ratio = evals.max() / np.sum(evals_smaller)

    return np.float64(stability_ratio)


__all__.append('stability_ratio')



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import warnings

import numpy as np

from .anam import AbstractAnam, register_class
from .stats import find_cov_multitrial
from .diags import ModelDiagnostics
from .utils import get_valid_samples

__all__ = []


def A_to_companion(A):
    """Transforms a [nsignals x nsignals x order] MVAR parameter array into the
    [nsignals*order x nsignals*order] companion form. This assumes there is no
    leading identity matrix in the A form.

    Parameters
    ----------
    A : ndarray
        MVAR parameter array of size [nsignals, nsignals, order]

    Returns
    -------
    ndarray
        MVAR parameter matrix in companion form

    """

    nsignals = A.shape[0]
    order = A.shape[2]
    companion = np.eye(nsignals*(order), k=nsignals, dtype=A.dtype).T
    companion[:nsignals, :] = A.reshape((nsignals, nsignals*order), order='F')

    return companion


def get_residuals(data, parameters, delay_vect, backwards=False, mode='valid'):
    """This is a helper function for computing the residuals of a dataset after
    the MVAR predictions have been removed.

    Parameters
    ----------
    data : ndarray
        Data to compute the residuals from, of size [nchannels x nsampes x nrealisations]
    parameters : ndarray
        MVAR parameter matrix, of size [nchannels x nchannels x model order]
    delay_vect : ndarray
        Vector of lag indices corresponding to the third dimension of the parameter matrix
    backwards : bool
        Flag indicating whether the forwards or backwards parameters havebeeen
        passed (Default value = False)
    mode : {'valid','full_nan','full'}
        Options for excluding or replacing residuals which do not have a full
        model prediction ie the third sample of an order 5 model. 'valid'
        removes samples without a full model prediction, 'full_nan' returns
        resids of the same size as the data with nans replacing excluded
        samples and 'full' returns resids keeping non-full model samples in
        place.

    Returns
    -------
    ndarray
        Residual data, of size [nchannels x nsamples x nrealisations]

    """
    # TODO: double check that we are including all the order
    resid = np.zeros_like(data, dtype=data.dtype)

    # for each epoch...
    for ntrial in range(data.shape[2]):
        if backwards:
            pred = np.zeros_like(data[:, :, 0], dtype=data.dtype)

            for m in range(1, len(delay_vect)):
                tmp = parameters[:, :, m].dot(data[:, :, ntrial])
                pred[:, :-delay_vect[m]] += tmp[:, delay_vect[m]:]

            resid[:, :, ntrial] = data[:, :, ntrial] - pred[...]

        else:
            pred = np.zeros_like(data[:, :, 0], dtype=data.dtype)

            for m in range(1, len(delay_vect)):
                tmp = parameters[:, :, m].dot(data[:, :, ntrial])
                pred[:, delay_vect[m]:] += tmp[:, :-delay_vect[m]]

            resid[:, :, ntrial] = data[:, :, ntrial] - pred[...]

    resid = get_valid_samples(resid, delay_vect, mode=mode)

    return resid


__all__.append('get_residuals')


class AbstractLinearModel(AbstractAnam):
    """This is a base class defining data storage and generic methods for
    LinearModel classes. New LinearModels should inherit from
    AbstractLinearModel and overload the fit_model method"""

    # This is only used within Anamnesis
    hdf5_outputs = ['data_cov', 'resid_cov', 'parameters',
                    'maxorder', 'delay_vect']

    def __init__(self):
        AbstractAnam.__init__(self)

    @property
    def nsignals(self):
        """Number of signals in fitted model"""
        return self.parameters.shape[0]

    @property
    def order(self):
        """Order of fitted model"""
        return self.parameters.shape[2] - 1

    @property
    def companion(self):
        """Parameter matrix in companion form"""
        return A_to_companion(self.parameters[:, :, 1:])

    def compute_diagnostics(self, data, compute_pc=False):
        """Compute several diagnostic metrics from a fitted MVAR model and a
        dataset.

        Parameters
        ----------
        data : ndarrat
            The data to compute the model diagnostics with, typically the same
            data used during model fit.
        compute_pc : bool
             Flag indicating whether the compute the percent consistency, this
             is typically the the longest computation of all the metrics here
             (Default value = False)

        Returns
        -------
        sails.ModelDiagnostics
            object containing LL, AIC, BIC, stability
            ratio, durbin-watson, R squared and percent consistency
            measures

        """

        from .diags import ModelDiagnostics

        return ModelDiagnostics.compute(self, data, compute_pc=compute_pc)

    def get_residuals(self, data, mode='valid'):
        """Returns the prediction error from a fitted model. This is a wrapper
        function for get_residuals()

        Parameters
        ----------
        data : ndarray
            Data to compute the residuals of, of size [nchannels x nsamples x nrealisations]

        Returns
        -------
        ndarray
            Residual data

        """

        if self.parameters.ndim == 3:
            resid = get_residuals(data,
                                  self.parameters,
                                  self.delay_vect,
                                  backwards=False,
                                  mode=mode)
        elif self.parameters.ndim == 4:
            # We have different parameters for each epoch
            resid = get_residuals(np.atleast_3d(data[..., 0]),
                                  self.parameters[..., 0],
                                  self.delay_vect,
                                  backwards=False,
                                  mode=mode)
            for e in range(1, self.parameters.shape[3]):
                tmp = get_residuals(np.atleast_3d(data[..., e]),
                                    self.parameters[..., e],
                                    self.delay_vect,
                                    backwards=False,
                                    mode=mode)
                resid = np.concatenate((resid, tmp), axis=2)

        return resid


__all__.append('AbstractLinearModel')
register_class(AbstractLinearModel)


class VieiraMorfLinearModel(AbstractLinearModel):
    """A class implementing the Vieira-Morf linear model fit"""

    # This is only used within Anamnesis
    hdf5_outputs = AbstractLinearModel.hdf5_outputs + ['bwd_parameters']

    def __init__(self):
        AbstractLinearModel.__init__(self)

    def get_residuals(self, data, forward_parameters=True, mode='valid'):
        """Returns the prediction error from a fitted model. This is a wrapper
        function for get_residuals()

        Parameters
        ----------
        data : ndarray
            Data to compute the residuals from, of size [nchannels x nsamples x nrealisations]
        forward_parameters : bool
            If True, use forward parameters, otherwise
            use backward parameters (Default value = True)

        Returns
        -------
        ndarray
            Residual data

        """

        if not forward_parameters:
            if self.bwd_parameters.ndim == 3:
                resid = get_residuals(data,
                                      self.bwd_parameters,
                                      self.delay_vect,
                                      backwards=True,
                                      mode=mode)
            elif self.bwd_parameters.ndim == 4:
                # We have different parameters for each epoch
                resid = get_residuals(data[..., 0],
                                      self.bwd_parameters[..., 0],
                                      self.delay_vect,
                                      backwards=True,
                                      mode=mode)
                for e in range(1, self.bwd_parameters.shape[3]):
                    tmp = get_residuals(data[..., e],
                                        self.bwd_parameters[..., e],
                                        self.delay_vect,
                                        backwards=True,
                                        mode=mode)
                    resid = np.concatenate((resid, tmp), axis=2)
        else:
            resid = AbstractLinearModel.get_residuals(self, data, mode=mode)

        return resid

    @classmethod
    def fit_model(cls, data, delay_vect):
        """Estimates the multichannel autoregressive spectral estimators
        using a Vieira-Morf algorithm. Equations are referenced to
        [Marple1987]_, appendix 15.B.

        This is the multitrial versions of the algorithm, using the AMVAR
        method outlined in [Ding2000]_.

        Parameters
        ----------
        data : numpy.ndarray
            array of shape [nsignals, nsamples, ntrials]
        delay_vect : numpy.ndarray
            Vector containing evenly spaced delays to be
            assessed in the model.  Delays are represented in
            samples.  Must start with zero.

        Returns
        -------
        sails.VieiraMorfLinearModel
            A populated object containing the fitted forward and
            backwards coefficients and several other useful variables and
            methods.

        """

        # Create object
        ret = cls()

        # Shorter reference
        X = data

        # Set-up initial parameters
        delay_vect = delay_vect.astype(int)
        nsignals = X.shape[0]
        ntrials = X.shape[2]
        maxorder = delay_vect.shape[0]-1

        # Begin model fitting
        #  Relevant parameters:
        #  A      - Forward linear prediction coefficients matrix
        #  B      - Backward linear prediction coefficients matrix
        #  PF     - Forward linear prediction error covariance
        #  PB     - Backward linear prediction error covariance
        #  PFH    - Estimate of forward linear prediction error covariance
        #  PBH    - Estimate of backward linear prediction error covariance
        #  PFBH   - Estimate of linear prediction error covariance
        #  RHO    - Estimate of the partial correlation matrix
        #  EF     - Forward linear prediction error
        #  EB     - Backward linear prediction error

        # Initialisation

        # Eq 15.91
        # Initialise prediction error as the data
        EF = X.copy().astype(X.dtype)
        EB = X.copy().astype(X.dtype)

        # Eq 15.90
        # Initialise error covariance as the data covariance
        PF = find_cov_multitrial(X, X)
        PB = find_cov_multitrial(X, X)

        # Set order zero coefficients to identity
        A = np.ndarray((nsignals, nsignals, maxorder+1), dtype=X.dtype)
        A[:, :, 0] = np.eye(nsignals)

        # TODO: double check that this is correct behaviour
        B = np.ndarray((nsignals, nsignals, maxorder+1), dtype=X.dtype)
        B[:, :, 0] = np.zeros(nsignals)

        # Main Loop

        M = 0
        while M < maxorder:

            # Create clean arrays for the estimates of the error covariance
            PFH = np.zeros((nsignals, nsignals))
            PBH = np.zeros((nsignals, nsignals))
            PFBH = np.zeros((nsignals, nsignals))

            # Eq 15.89 - get estimates of forward and backwards covariance
            PFH = find_cov_multitrial(EF[:, delay_vect[M+1]:, :],
                                      EF[:, delay_vect[M+1]:, :])
            PBH = find_cov_multitrial(EB[:, delay_vect[M]:-delay_vect[1], :],
                                      EB[:, delay_vect[M]:-delay_vect[1], :])
            PFBH = find_cov_multitrial(EF[:, delay_vect[M+1]:, :],
                                       EB[:, delay_vect[M]:-delay_vect[1], :])

            # Eq 15.88 - compute estimated normalised partial correlation
            # matrix
            tmp = np.linalg.inv(np.linalg.cholesky(PFH)).dot(PFBH)
            chol = np.linalg.cholesky(PBH)
            RHO = tmp.dot(np.array(np.mat(np.linalg.inv(chol)).H))

            M += 1

            # Eq 15.82 & 15.83 - Update forward and backward reflection
            # coefficients
            tmp = np.linalg.cholesky(PF).dot(RHO)
            A[:, :, M] = - tmp.dot(np.linalg.inv(np.linalg.cholesky(PB)))

            tmp = np.linalg.cholesky(PB).dot(np.array(np.mat(RHO).H))
            B[:, :, M] = - tmp.dot(np.linalg.inv(np.linalg.cholesky(PF)))

            # Eq 15.75 & 15.76 - Update forward and backward error covariances
            tmp = np.eye(nsignals) - (A[:, :, M].dot(B[:, :, M]))
            PF = tmp.dot(PF)

            tmp = np.eye(nsignals) - (B[:, :, M].dot(A[:, :, M]))
            PB = tmp.dot(PB)

            # might not need this if statement, subsequent xrange will handle
            # it, # though this is clearer
            if M > 1:
                A_t = A.copy()
                B_t = B.copy()

                # We are working from the equations - the FORTRAN is wrong!

                # Eq 15.71 & 15.72 - Update forward and backward predictor
                # coefficients
                for k in range(1, M):
                    mk = M-k

                    # TODO  - double check backwards coeffs against Ding2000
                    # simulations
                    A[:, :, k] = A_t[:, :, k] + A[:, :, M].dot(B_t[:, :, mk])
                    B[:, :, k] = B_t[:, :, k] + B[:, :, M].dot(A_t[:, :, mk])

            # Eq 15.84 & 15.85 - Update the residuals
            for t in range(ntrials):
                tmp = EF[:, :, t].copy()
                tmp2 = EB[:, :, t].copy()

                update_A = A[:, :, M].dot(tmp2[:, :-delay_vect[1]])
                EF[:, delay_vect[1]:, t] = tmp[:, delay_vect[1]:] + update_A

                update_B = B[:, :, M].dot(tmp[:, delay_vect[1]:])
                EB[:, delay_vect[1]:, t] = tmp2[:, :-delay_vect[1]] + update_B

        # Sanity check for single channel case
        if nsignals == 1 and np.allclose(PF, PB) is False:
            ret.status = -1
            warnings.warn("Warning: problem with model fit. PF != PB. "
                          "See single channel identity in Marple "
                          "top of p405")

        # Correct for sign flip in VM estimatation
        ret.parameters = -A
        ret.bwd_parameters = -B

        # Populate return object
        ret._data = X
        ret.maxorder = maxorder
        ret.delay_vect = delay_vect
        ret.data_cov = find_cov_multitrial(X, X)
        resids = ret.get_residuals(X)
        ret.resid_cov = find_cov_multitrial(resids, resids)

        return ret


__all__.append('VieiraMorfLinearModel')
register_class(VieiraMorfLinearModel)


class OLSLinearModel(AbstractLinearModel):
    """A class implementing ordinary least squares linear model fit"""

    @classmethod
    def fit_model(cls, data, delay_vect, estimator=None):
        """This is a class method which fits a linear model and returns a
        populated :class:`OLSLinearModel` instance containing the fitted model.

        Parameters
        ----------
        data : ndarray
            The data to be used to compute the model fit
        delay_vect : ndarray
            A vector of lag indices defining the lags to fit
        estimator : None or sklearn class
            If None, fit using standard OLS normal equations.
            If set to an appropriate sklearn class, use that estimator.


        Returns
        -------
        sails.OLSLinearModel
            A populated object containing the fitted coefficients and several
            other useful variables and methods.

        """

        # Create object
        ret = cls()

        # Set-up initial parameters
        nsignals = data.shape[0]
        nsamples = data.shape[1]
        maxorder = delay_vect.shape[0]-1
        maxlag = delay_vect[-1].astype(int)

        # Preallocate design matrix
        X = np.zeros((nsamples-maxlag, nsignals*maxorder))
        # Preallocate observation matrix
        Y = np.zeros((nsamples-maxlag, nsignals))

        # Create design matrix
        for idx in range(nsamples-maxlag):
            X[idx, :] = data[:, -delay_vect[1::].astype(int)+idx].reshape(-1)
            Y[idx, :] = data[:, idx, 0]

        # B is shaped [maxorder*nsignals, nsignals]
        #
        # [ A[1, 1, 1] A[1, 1, 2] ... A[1, 1, p] A[1, 2, 1] A[1, 2, 2] ... A[1, 2, p] ]
        # [ A[2, 1, 1] A[2, 1, 2] ... A[2, 1, p] A[2, 2, 1] A[2, 2, 2] ... A[2, 2, p] ]

        if estimator is None:
            # Use normal equations
            B = np.linalg.inv(X.T.dot(X)).dot(X.T).dot(Y)
        else:
            # Fit the passed in estimator with our data
            estimator.fit(X, Y)

            # The coefficients are transposed in the sklearn implementation
            B = estimator.coef_.T

        # Reshape output
        parameters = np.zeros((nsignals, nsignals, maxorder+1))
        parameters[..., 0] = -np.eye(nsignals)
        for idx in range(nsignals):
            parameters[:, idx, 1:] = B[idx*maxorder:(idx+1)*maxorder].T

        resids = get_residuals(data, -parameters, delay_vect)
        ret.data_cov = find_cov_multitrial(data, data)
        ret.resid_cov = find_cov_multitrial(resids, resids)
        ret.parameters = parameters
        ret.maxorder = maxorder
        ret.delay_vect = delay_vect

        return ret


__all__.append('OLSLinearModel')
register_class(OLSLinearModel)


def sliding_window_fit(model_class, data, delay_vect,
                       win_len_samp, win_step_samp,
                       compute_diagnostics=True):
    """A helper function for fitting many MVAR models within sliding windows
    across a dataset.

    Parameters
    ----------
    model_class : LinearModel
        The SAILS linear model typee to fit
    data : ndarray
        Data to fit the model on, of size [nchannels x nsamples x nrealisations]
    delay_vect : ndarray
        A vector of lags specifying which lags to fit
    win_len_samp : int
        The window length in samples
    win_step_samp : int
        The step size between windows in samples
    compute_diagnostics : boolean
        Flag indicating whether to compute model diagnostics at each window
        (Default value = True)

    Returns
    -------
    LinearModel
        An instance of linear model passed into the function containing MVAR
        parameters for all windows. The parameters are stored in
        model.parameters which is of size:
        [nchannels x nchannels x model order x nwindows]

    """

    # Compute start and end samples of our windows
    start_pts = list(range(0,
                           data.shape[1] - max(delay_vect) - win_len_samp,
                           win_step_samp))
    end_pts = [x + win_len_samp for x in start_pts]

    # Preallocate model outputs
    params = np.zeros((data.shape[0], data.shape[0],
                       len(delay_vect), len(start_pts)))
    resid_cov = np.zeros((data.shape[0], data.shape[0],
                          len(start_pts)))

    # Preallocate diagnostic outputs
    if compute_diagnostics:
        diag = []

    # Main model loop
    x = np.zeros((win_len_samp, len(start_pts)))
    for ii in range(len(start_pts)):
        x = data[:, start_pts[ii]:end_pts[ii], :]

        # Fit model
        m = model_class.fit_model(x, delay_vect)
        params[:, :, :, ii] = m.parameters
        resid_cov[:, :, ii] = m.resid_cov

        # Compute diagnostics
        if compute_diagnostics:
            diag.append(m.compute_diagnostics(x))

    # Create output class
    M = model_class()
    M.parameters = params
    M.resid_cov = resid_cov
    M.delay_vect = delay_vect
    M.win_len_samp = win_len_samp
    M.win_step_samp = win_step_samp
    M.time_vect = (np.array(start_pts) + np.array(end_pts)) / 2

    if compute_diagnostics:
        D = ModelDiagnostics.combine_diag_list(diag)

    if compute_diagnostics:
        return M, D
    else:
        return M


__all__.append('sliding_window_fit')


def pca_reduced_fit(X, delay_vect, ndim, linear_model=VieiraMorfLinearModel):
    """
    Helper for computing an MVAR on dimensionality reduced data (using PCA).

    Returns a
    model fitted to reduced data, the reduced model projected back to original
    data dimensions and the pca object used for reduction

    Parameters
    ----------
    X : ndarray
        The data to compute the reduced model fit on
    delay_vect : ndarray
        A vector of lags specifying which lags to fit
    ndim : int
        Number of components to reduce data to.
    linear_model : class
        Subclass of AbstractLinearModel to use for fit.
        Defaults to VieiraMorfLinearModel.
        (Default value = True)

    Returns
    -------
    red_model : AbstractLinearModel
        An instance of the linear model passed into the function containing MVAR
        parameters for all windows for the reduced model.

    proj_model : AbstractLinearModel
        An instance of the linear model passed into the function containing MVAR
        parameters for all windows for the projected model.

    pc : sails.utils.PCA
        PCA class with information about the PCA projection included
    """

    # PCA requires [samples x channels] whereas data is [channels x samples]
    from .utils import PCA

    pc = PCA(X.T, ndim)
    Xred = pc.scores.T
    Xred = Xred[:, :, None]

    # Fit MVAR to dim-reduced data
    red_model = linear_model.fit_model(Xred, delay_vect)

    # Project back to data dimensions
    Afull = np.zeros((X.shape[0], X.shape[0], len(delay_vect)))

    # Don't project identity at start
    Afull[:, :, 0] = -np.eye(X.shape[0])

    for ii in range(1, len(delay_vect)):
        Afull[:, :, ii] = np.dot(pc.components.T, red_model.parameters[:, :, ii]).dot(pc.components)

    resid_cov_full = np.dot(pc.components.T, red_model.resid_cov[:, :]).dot(pc.components)[:, :, None]

    # Store projected model
    proj_model = linear_model()
    proj_model.parameters = Afull
    proj_model.resid_cov = resid_cov_full
    proj_model.delay_vect = delay_vect

    return red_model, proj_model, pc


__all__.append('pca_reduced_fit')


def coloured_noise_fit(X, model_order):
    """A helper function to fit an autoregressive model whose parameters are
    constrained to be coloured noise as definede by [Kasdin1995]_.

    Parameters
    ----------
    X : ndarray
        The data to compute the model fit on
    model_order : int
        The maximum lag to be computed

    Returns
    -------
    LinearModel
        A SAILS linear model instance containing the fitted model
    float
        The power of the final model fit.

    """

    from .tutorial_utils import generate_pink_roots, model_from_roots

    # Initialise returned model
    M = AbstractLinearModel()
    M.parameters = np.zeros((X.shape[0], X.shape[0], model_order, X.shape[2]))
    M.resid_cov = np.eye(X.shape[0])  # Assumed for now
    M.delay_vect = np.arange(model_order)

    # Range of 1/f^alpha values to explore
    powers = np.linspace(.05, 1.95, 200)

    fit_inds = np.zeros((X.shape[0],), dtype=int)
    for ichan in range(X.shape[0]):
        ss = np.zeros(powers.shape)
        for ii in range(len(powers)):

            rts = generate_pink_roots(powers[ii], order=model_order)
            m = model_from_roots(rts)

            resid = get_residuals(X[None, ichan, :, :], m.parameters, np.arange(model_order))
            ss[ii] = np.power(resid, 2).sum()

        fit_inds[ichan] = np.argmin(ss).astype(int)
        rts = generate_pink_roots(powers[fit_inds[ichan]], order=model_order)
        M.parameters[ichan, ichan, :, 0] = -np.poly(rts)

    fit_powers = powers[fit_inds]

    return M, fit_powers


__all__.append('coloured_noise_fit')



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import unittest

import numpy as np

from ..support import array_assert

__all__ = []


class test_model_fits(unittest.TestCase):

    def run_fit_checks(self, freq, fitfunc, extraargs=None):
        from ..modal import MvarModalDecomposition

        if extraargs is None:
            extraargs = {}

        sample_rate = 100
        seconds = 50
        time_vect = np.linspace(0, seconds, sample_rate * seconds)
        x = np.sin(2*np.pi*freq*time_vect)

        model = fitfunc.fit_model(x[None, :, None], np.arange(3), **extraargs)
        modes = MvarModalDecomposition.initialise(model, sample_rate=sample_rate)

        # Compute analytic modes
        mode_theta = (freq/(sample_rate/2.0)) * np.pi
        mode_poles = np.array([np.cos(mode_theta) + 1j*np.sin(mode_theta),
                               np.cos(mode_theta) - 1j*np.sin(mode_theta)])

        # Check fitted parameters are close to analytic parameters
        array_assert(-np.poly(mode_poles),
                     model.parameters[0, 0, :].real,
                     decimal=3)

        # Check fitted poles are close to analytic poles
        array_assert(mode_poles,
                     modes.evals[:, 0],
                     decimal=3)

    def test_ols(self):
        """Test the Ordinary Least Squares model fits"""

        from ..modelfit import OLSLinearModel
        fit = OLSLinearModel

        self.run_fit_checks(24, fit)
        self.run_fit_checks(10, fit)
        self.run_fit_checks(40, fit)

    def test_ols_sklearn(self):
        """Test the Ordinary Least Squares model fits using sklearn"""

        import sklearn.linear_model

        from ..modelfit import OLSLinearModel

        fit = OLSLinearModel
        estimator = sklearn.linear_model.LinearRegression(fit_intercept=False)
        extraargs = {'estimator': estimator}

        self.run_fit_checks(24, fit, extraargs=extraargs)
        self.run_fit_checks(10, fit, extraargs=extraargs)
        self.run_fit_checks(40, fit, extraargs=extraargs)

    def test_vieiramorf(self):
        """Test the Vieira-Morf model fits"""

        from ..modelfit import VieiraMorfLinearModel
        fit = VieiraMorfLinearModel

        self.run_fit_checks(24, fit)
        self.run_fit_checks(10, fit)
        self.run_fit_checks(40, fit)



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

# Avoid having code to make anamnesis optional in multiple places
try:
    from anamnesis import AbstractAnam, register_class
except ImportError:
    AbstractAnam = object

    def register_class(x):
        return None

__all__ = ['AbstractAnam', 'register_class']



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import numpy as np

from .anam import AbstractAnam, register_class
from sails.support import ensure_leading_positive, ensure_leading_negative
from sails import plotting

__all__ = []


class AbstractMVARMetrics(AbstractAnam):
    """This is a base class for computing various connectivity metrics from
    fitted autoregressive models"""

    hdf5_outputs = []

    def __init__(self):
        AbstractAnam.__init__(self)

    @property
    def S(self):
        """Spectral Matrix of shape [nsignals, nsignals, order, epochs]"""
        return spectral_matrix(self.H, self.resid_cov)

    @property
    def PSD(self):
        """Power Spectral Density matrix"""
        return np.abs(psd_matrix(self.S, self.sample_rate))

    @property
    def inv_S(self):
        """Inverse Power Spectrl Density matrix"""
        return inv_spectral_matrix(self.H, self.resid_cov)

    @property
    def coherency(self):
        """Complex-valued coherency"""
        return coherency(self.S)

    @property
    def magnitude_squared_coherence(self):
        """Magnitude squared coherence"""
        return magnitude_squared_coherence(self.S)

    @property
    def imaginary_coherence(self):
        """Imaginary part of complex coherency"""
        return imaginary_coherence(self.S)

    @property
    def phase_coherence(self):
        """Phase angle of complex coherency"""
        return phase_coherence(self.S)

    @property
    def partial_coherence(self):
        """Partialled magnitude-squared coherence"""
        return partial_coherence(self.inv_S)

    @property
    def ff_directed_transfer_function(self):
        """Full-frequency directed transfer function"""
        return ff_directed_transfer_function(self.H)

    @property
    def d_directed_transfer_function(self):
        """Direct directed transfer function"""
        return d_directed_transfer_function(self.H, self.inv_S)

    @property
    def directed_transfer_function(self):
        """Directed transfer function"""
        return directed_transfer_function(self.H)

    @property
    def partial_directed_coherence(self):
        """Partial directed coherence"""
        return partial_directed_coherence(self.Af)

    @property
    def isolated_effective_coherence(self):
        """Isolated effective coherence"""
        return isolated_effective_coherence(self.Af, self.resid_cov)

    @property
    def geweke_granger_causality(self):
        """Geweke granger causality"""
        return geweke_granger_causality(self.S, self.H, self.resid_cov)

    def plot_diags(self, metric='S', inds=None, F=None, ax=None):
        """Helper function for plotting the diagonal part of a connectivty
        metric. Calls sails.plotting.plot_diagonal.

        Parameters
        ----------
        metric : str
            Name of the connectivity metric to plot (Default value = 'S')
        F : matplotlib figure handle
            Handle for matplotlib figure to plot in (Default value = None)
        ax : matplotlib axes handle
            Handle for matplotlib axes to plot in (Default value = None)

        """

        met = getattr(self, metric)
        plotting.plot_diagonal(self.freq_vect, met, title=metric, F=F, ax=ax)

    def plot_summary(self, metric='S', ind=1):
        """Helper function for plotting a summary of a connectivty metric.
        Calls sails.plotting.plot_summary.

        Parameters
        ----------
        metric : str
            Name of the connectivity metric to plot (Default value = 'S')
        ind : int
           Index into frequency dimension to plot connectivity matrix (Default value = 1)


        """

        met = getattr(self, metric)
        plotting.plot_metric_summary(self.freq_vect, met, ind=ind)

    def get_spectral_generalisation(self, metric='S'):
        """Compute the spectral generalisation matrix for a given metric.
        Returns the spatial similarity of different frequencies.

        Parameters
        ----------
        metric : str
            Name of the connectivity metric to plot (Default value = 'S')

        Returns
        -------
        ndarray
            Matrix of spectral generalisation [nfrequencies x nfrequencies]

        """

        met = np.abs(getattr(self, metric)[:, :, :, 0])
        met = met.reshape((met.shape[0]**2, met.shape[2]))

        return met.T.dot(met)

    def get_node_weight(self, metric='S'):
        """Compute the node weight for a given connectivity metric.

        Parameters
        ----------
        metric : str
            Name of the connectivity metric to plot (Default value = 'S')

        Returns
        -------
        ndarray
            Node-weight matrix of size [nchannels x nchannels]


        """

        met = np.abs(getattr(self, metric)[:, :, :, 0])
        for ii in range(met.shape[2]):
            met[:, :, ii] = met[:, :, ii] - np.diag(np.diag(met[:, :, ii]))

        nconnections = met.shape[0]+met.shape[1]-1

        return met[:, 0, :] + met[0, :, :] / nconnections


__all__.append('AbstractMVARMetrics')
register_class(AbstractMVARMetrics)


class ModalMvarMetrics(AbstractMVARMetrics):
    """This class for computing various connectivity metrics based on a Modal
    decomposition of a fitted MVAR model

    This is typically called using the :meth:`initialise` or
    :meth:`initialise_from_modes` classmethods to compute the decomposition and
    return a class instance containing them.

    """

    def __init__(self):
        AbstractMVARMetrics.__init__(self)

    # This is only used when used within Anamnesis
    hdf5_outputs = ['resid_cov', 'freq_vect', 'sample_rate', 'H', 'modes']

    @classmethod
    def initialise(cls, model, sample_rate, freq_vect, sum_modes=True):
        """Compute some basic values from the pole-residue decomposition of the
        A matrix. This is a class method which creates and returns a
        :class:`ModalMvarMetrics` instance containing the decomposition
        parameters with methods computing various connectivity metrics.

        This class computes the transfer function H using modal parameters.

        Currently only implemented for a single realisation (A.ndim == 3)

        Parameters
        ----------
        model : sails LinearModel
            SAILS class instance containing a fitted linear model.
        sample_rate : float
            Sample rate of the dataset
        freq_vect : ndarray
            Vector of specifying which frequencies to compute connectivity at.
        sum_modes : bool
            Flag indicating whether to sum across modes (Default value = True)

        Returns
        -------
        sails.ModalMvarMetrics instance

        """

        # Create object
        ret = cls()
        A = model.parameters
        resid_cov = model.resid_cov

        if A.ndim == 4:
            NotImplementedError('Modal Metrics only implemented for single '
                                'realisations of A')
        else:
            # Add dummy epoch
            A = A[..., None]

        A = ensure_leading_negative(A)

        if resid_cov.ndim == 2:
            resid_cov = resid_cov[..., None]  # Add a dummy epoch
        if resid_cov.shape[2] != A.shape[0]*A.shape[2]:
            # Repeat resid cov per epoch/mode (probably only valid for
            # modes...)
            resid_cov = np.repeat(resid_cov, A.shape[0]*A.shape[2], axis=-1)

        ret.resid_cov = resid_cov
        ret.freq_vect = freq_vect
        ret.sample_rate = sample_rate

        # Compute transfer function
        from .modal import MvarModalDecomposition
        ret.modes = MvarModalDecomposition.initialise(model,
                                                      sample_rate=sample_rate,
                                                      normalise=False)
        ret.H = ret.modes.transfer_function(sample_rate, freq_vect, sum_modes=sum_modes)

        return ret

    @classmethod
    def initialise_from_modes(cls, model, modes, sample_rate, freq_vect, mode_inds=None, sum_modes=True):
        """Compute some basic values from the pole-residue decomposition of the
        A matrix. This is a class method which creates and returns a
        :class:`sails.ModalMvarMetrics` instance containing the decomposition
        parameters with methods computing various connectivity metrics.

        This class computes the transfer function H using modal parameters. A
        sub-set of modes can be used to compute a reduced H.

        Currently only implemented for a single realisation (A.ndim == 3)

        Parameters
        ----------
        model : sails LinearModel
            SAILS class instance containing a fitted linear model.
        modes : sails MvarModalDecomposition
            SAILS class instance containing a fitted modal decomposition
        sample_rate : float
            The sample rate of the fitted data
        freq_vect : ndarray
            Vector of specifying which frequencies to compute connectivity at.
        mode_inds : ndarray
            Which modes to use when constructing H (Default value = None)
        sum_modes :Boolean
            Flag indicating whether to sum across modes when constructing H
            (Default value = True)

        Returns
        -------
        sails.ModalMvarMetrics instance

        """
        # Create object
        ret = cls()
        ret.modes = modes
        ret.freq_vect = freq_vect
        ret.sample_rate = float(sample_rate)

        A = model.parameters
        resid_cov = model.resid_cov

        if A.ndim == 4:
            NotImplementedError('Modal Metrics only implemented for single '
                                'realisations of A')
        else:
            # Add dummy epoch
            A = A[..., None]

        A = ensure_leading_positive(A)

        if resid_cov.ndim == 2:
            resid_cov = resid_cov[..., None]  # Add a dummy epoch
        if resid_cov.shape[2] != A.shape[0]*A.shape[2]:
            # Repeat resid cov per epoch/mode (probably only valid for
            # modes...)
            resid_cov = np.repeat(resid_cov, A.shape[0]*A.shape[2], axis=-1)
        ret.resid_cov = resid_cov

        if mode_inds is None:
            mode_inds = np.arange(modes.nmodes)

        # Compute transfer function
        ret.H = ret.modes.transfer_function(sample_rate, freq_vect,
                                            modes=mode_inds,
                                            sum_modes=sum_modes)

        return ret


__all__.append('ModalMvarMetrics')
register_class(ModalMvarMetrics)


class FourierMvarMetrics(AbstractMVARMetrics):
    """This class for computing various connectivity metrics based on the
    Fourier transform of the MVAR co-efficients.

    This is typically called using the :meth:`initialise` classmethod to compute
    the frequency transform and return a class instance with methods for
    computing different connectivity estimators

    """

    # This is only used when used within anamnesis
    hdf5_outputs = ['resid_cov', 'freq_vect', 'sample_rate', 'H', '_A', 'Af']

    def __init__(self):
        AbstractMVARMetrics.__init__(self)

    @classmethod
    def initialise(cls, model, sample_rate, freq_vect, nmodes=None):
        """Compute some basic values from the Fourier transform of the A
        matrix. This is a class method which creates and returns a
        :class:`FourierMvarMetrics` instance with methods computing
        various connectivity metrics.

        This class computes the transfer function H the Fourier transform.

        Parameters
        ----------
        model : sails LinearModel
            SAILS class instance containing a fitted linear model.
        sample_rate : float
            Sample rate of the dataset
        freq_vect : ndarray
            Vector of specifying which frequencies to compute connectivity at.

        Returns
        -------
        sails.FourierMvarMetrics instance

        """

        # Create object
        ret = cls()
        A = model.parameters
        resid_cov = model.resid_cov

        if A.ndim == 3:
            A = A[..., None]  # Add a dummy epoch

        ret._A = A

        A = ensure_leading_positive(A)

        if resid_cov.ndim == 2:
            resid_cov = resid_cov[..., None]  # Add a dummy epoch

        if resid_cov.shape[2] != A.shape[3]:
            # Repeat resid cov per epoch/mode (probably only valid for
            # modes...)
            resid_cov = np.repeat(resid_cov, A.shape[3], axis=-1)

        ret.resid_cov = resid_cov
        ret.freq_vect = freq_vect
        ret.sample_rate = float(sample_rate)

        # Get frequency transform of A
        ret.Af = ar_spectrum(A, resid_cov, sample_rate, freq_vect)

        # Get transfer function
        ret.H = transfer_function(ret.Af)

        return ret


__all__.append('FourierMvarMetrics')
register_class(FourierMvarMetrics)


def modal_transfer_function(evals, evecl, evecr, nchannels,
                            sample_rate=None, freq_vect=None):
    """Compute the transfer function in pole-residue form, splitting the system
    into modes with separate transfer functions. The full system transfer
    function is then a linear sum of each modal transfer function.

    Parameters
    ----------
    evals : ndarray
        Complex valued eigenvalues from an eigenvalue decomposition of an MVAR
        parameter matrix.
    evecl : ndarray
        Complex valued left-eigenvectorsfrom an eigenvalue decomposition of an
        MVAR parameter matrix.
    evecr : ndarry
        Complex valued right-eigenvectorsfrom an eigenvalue decomposition of an
        MVAR parameter matrix.
    nchannels : int
        Number of channels in the decomposed system
    sample_rate : float
        The sampling  rate of the decomposed system (Default value = None)
    freq_vect : ndarray
        Vector of frequencies at which to evaluate the transfer function (Default value = None)

    Returns
    -------
    ndarray
        Modal transfer-function (H) of size [nchannels x nchannels x nfrequencies x nmodes]

    """

    if (sample_rate is None and freq_vect is not None) or \
       (sample_rate is not None and freq_vect is None):
        raise ValueError('Please define both sample_rate and freq_vect '
                         '(or neither to return normalised frequency 0->pi)')

    if freq_vect is None and sample_rate is None:
        # Use normalised frequency
        freq_vect_rads = np.linspace(0, np.pi, 512)
    else:
        # convert user defined frequecies to radians
        freq_vect_rads = (freq_vect / (sample_rate / 2.)) * np.pi

    # Compute transfer function per mode
    H = np.zeros((nchannels, nchannels,
                 len(freq_vect_rads), len(evals)), dtype=complex)

    # Compute point on unit circle for each frequency
    z = np.cos(freq_vect_rads) + 1j*np.sin(freq_vect_rads)

    for imode in range(len(evals)):

        # Compute residue matrix from left and right eigenvectors
        l = np.atleast_2d(evecl[:nchannels, imode]).T
        r = np.atleast_2d(evecr[:nchannels, imode]).conj()
        residue = l.dot(r)

        for ifreq in range(len(freq_vect_rads)):

            # Compute transfer function for this frequency and this mode
            num = residue * z[ifreq]
            dom = z[ifreq] - evals[imode]
            H[:, :, ifreq, imode] = num / dom

    return H


__all__.append('modal_transfer_function')


# Spectrum estimators


def sdf_spectrum(A, sigma, sample_rate, freq_vect):
    """Estimate of the Spectral Density as found on wikipedia and [Quirk1983]_.

    This assumes that the spectral representation of A is invertable

    Parameters
    ----------
    A : ndarray
        Matrix of autoregressive parameters, of size [nchannels x nchannels x model order]
    sigma : ndarray
        Residual covariance matrix of the modelled system
    sample_rate : float
        The samplingfrequency of the modelled system
    freq_vect : ndarray
        Vector of frequencies at which to evaluate the spectrum

    Returns
    -------
    ndarray
        Frequenecy transform of input A, of size [nchannels x nchannels x nfrequencies]


    """
    # model order
    order = A.shape[2] - 1
    nsignals = A.shape[0]
    T = 1./sample_rate
    spectrum = np.zeros((nsignals, nsignals, len(freq_vect)), dtype=complex)
    eye = np.eye(nsignals)

    for node1 in range(nsignals):
        for node2 in range(nsignals):
            for idx, f in zip(range(len(freq_vect)), freq_vect):

                est = eye[node1, node2]

                for k in range(1, order+1):
                    est = A[node1, node2, k] * np.exp(0-1j*2*np.pi*k*f*T)

                num = sigma[node1, node2] * T
                dom = np.abs(est)**2
                spectrum[node1, node2, idx] = num / dom

    return spectrum


__all__.append('sdf_spectrum')


def psd_spectrum(A, sigma, sample_rate, freq_vect):
    """Estimate the PSD representation of a set of MVAR coefficients as stated in
    [Penny2009]_ section 7.4.4.

    This assumes that the spectral representation of A is invertable

    WARNING: does not behave as expected for some data, use with caution.
    ar_spectrum is recommended.

    Parameters
    ----------
    A : ndarray
        Matrix of autoregressive parameters, of size [nchannels x nchannels x model order]
    sigma : ndarray
        Residual covariance matrix of the modelled system
    sample_rate : float
        The samplingfrequency of the modelled system
    freq_vect : ndarray
        Vector of frequencies at which to evaluate the spectrum

    Returns
    -------
    ndarray
        Frequenecy transform of input A, of size [nchannels x nchannels x nfrequencies]


    """
    # model order
    order = A.shape[2] - 1
    nsignals = A.shape[0]
    T = 1./sample_rate
    spectrum = np.zeros((nsignals, nsignals, len(freq_vect)), dtype=complex)
    eye = np.eye(nsignals)

    for node1 in range(nsignals):
        for node2 in range(nsignals):
            for idx, f in zip(range(len(freq_vect)), freq_vect):

                # Normalised digital frequency
                f = f / sample_rate

                est = eye[node1, node2]

                for k in range(1, order+1):
                    est += A[node1, node2, k] * np.exp(0-1j*2*np.pi*k*f)

                spectrum[node1, node2, idx] = est

    # Compute the PSD_mvar

    psd = np.zeros_like(spectrum)
    for idx, f in zip(range(len(freq_vect)), freq_vect):
        A_inv = np.linalg.inv(spectrum[..., idx])
        # Hermitian of the inverse
        A_inv_herm = np.array(np.mat(A_inv).H)
        psd[:, :, idx] = T * (A_inv.dot(sigma).dot(A_inv_herm))

    return psd


__all__.append('psd_spectrum')


def ar_spectrum(A, sigma, sample_rate, freq_vect):
    """Estimate the spectral representation of a set of MVAR coefficients as
    suggested by [Baccala2001]_, the equation without a number just
    below equation 13.

    Parameters
    ----------
    A : ndarray
        Matrix of autoregressive parameters, of size [nchannels x nchannels x model order]
    sigma : ndarray
        Residual covariance matrix of the modelled system
    sample_rate : float
        The samplingfrequency of the modelled system
    freq_vect : ndarray
        Vector of frequencies at which to evaluate the spectrum

    Returns
    -------
    ndarray
        Frequenecy transform of input A, of size [nchannels x nchannels x nfrequencies]


    """

    # This routine assumes that our parameter matrix is of the form where
    # we have a leading positive.  Ensure that this is the case.
    A = ensure_leading_positive(A)

    if A.ndim == 3:
        A = A[..., None]

    # model order
    order = A.shape[2] - 1

    nsignals = A.shape[0]
    nepochs = A.shape[3]

    spectrum = np.zeros((nsignals, nsignals,
                         len(freq_vect), nepochs), dtype=complex)

    for idx, f in zip(range(len(freq_vect)), freq_vect):
        # Normalised digital frequency
        f = f / sample_rate

        est = 0+0j

        for k in range(1, order+1):
            est -= A[:, :, k, :] * np.exp(0-1j*2*np.pi*k*f)

        spectrum[:, :, idx, :] = est

    return spectrum


__all__.append('ar_spectrum')


def transfer_function(Af):
    """Function for computing the transfer function of a system from the
    frequency transform of the autoregressive parameters.

    Parameters
    ----------
    Af :
        Frequency domain version of parameters (can be calculated using
        ar_spectrum function) [nchannels x nchannels x nfreqs x nepochs]

    Returns
    -------
    ndarray
        System transfer function, of size [nchannels x nchannels x nfreqs x nepochs]


    Note
    ----

    This function computes this the transfer function :math:`H` as defined by:

    .. math:: H(f) = ( I - Af(f) )^{-1}

    where :math:`Af` is the frequency transformed autoregessive parameter
    matrix and :math:`I` is identity.

    """

    # Remove identity from Af
    Afi = np.repeat(np.eye(Af.shape[0])[:, :, None],
                    Af.shape[2], axis=-1)[..., None] - Af

    nchannels = Af.shape[0]
    nfreqs = Af.shape[2]
    nepochs = Af.shape[-1]

    tf = np.zeros((nchannels, nchannels,
                   nfreqs, nepochs), dtype=complex)

    for e in range(nepochs):
        for f in range(nfreqs):
            tf[:, :, f, e] = np.linalg.inv(Afi[:, :, f, e])

    return tf


__all__.append('transfer_function')


def spectral_matrix(H, noise_cov):
    """Function for computing the spectral matrix, the matrix of spectra and
    cross-spectra.

    Parameters
    ----------
    H : ndarray
        The transfer matrix [nchannels x nchannels x nfreqs x nepochs]
    noise_cov : ndarray
        The noise covariance matrix of the system
        [nchannels x nchannels x nepochs]

    Returns
    -------
    ndarray
        System spectral matrix, of size [nchannels x nchannels x nfreqs x nepochs]

    Note
    ----

    This function computes the system spectral matrix as defined by:

    .. math:: S(f) = H(f)\\Sigma H(f)^H

    where :math:`H` is the  transfer function, :math:`\\Sigma` is the residual
    covariance and :math:`^H` the Hermitian transpose.

    """

    # Make sure that we have the 4th dimension
    if H.ndim < 4:
        H = H[:, :, :, None]

    S = np.zeros_like(H)

    for i in range(H.shape[2]):
        for j in range(H.shape[3]):
            S[:, :, i, j] = H[:, :, i, j].dot(
                noise_cov[:, :, j]).dot(H[:, :, i, j].T.conj())

    return S


__all__.append('spectral_matrix')


def psd_matrix(S, sample_rate):
    """Function for computing the power spectral density matrix with units matched
    to scipy.signal.welch(X, axis=1, scaling='spectrum')

    Parameters
    ----------
    S :
        The spectral matrix [nchannels x nchannels x nfreqs x nepochs]
    sample_rate :
        The sampling rate of the system

    Returns
    -------


    """

    return S / sample_rate


__all__.append('psd_matrix')


def inv_spectral_matrix(H, noise_cov):
    """Function for computing the inverse spectral matrix.

    Parameters
    ----------
    H :
        The transfer matrix [nchannels x nchannels x nfreqs x nepochs]
    noise_cov :
        The noise covariance matrix of the system
        [nchannels x nchannels x nepochs]

    Returns
    -------
    ndarray
        Inverse system spectral matrix, of size [nchannels x nchannels x nfreqs x nepochs]

    Note
    ----

    This function computes the inverse spectral matrix as defined by:

    .. math:: S(f) = ( H(f)\\Sigma H(f)^H )^{-1}

    where :math:`H` is the  transfer function, :math:`\\Sigma` is the residual
    covariance and :math:`^H` the Hermitian transpose.

    """

    inv_S = np.zeros_like(H)

    S = spectral_matrix(H, noise_cov)

    for f in range(S.shape[2]):
        for e in range(S.shape[3]):
            inv_S[:, :, f, e] = np.linalg.pinv(np.abs(S[:, :, f, e]))

    return inv_S


__all__.append('inv_spectral_matrix')


def coherency(S):
    """Method for computing the Coherency. This is the complex form of
    coherence, from which metrics such as magnitude squared coherence can be
    derived

    Parameters
    ----------
    S : ndarray
        The system spectral matrix in 3D or 4D form, of size
        [nchannels x nchannels x nfreqs x nepochs]

    Returns
    -------
    ndarray
        Complex-valued coherency matrix, of size [nchannels x nchannels x nfreqs x nepochs]

    Note
    ----

    This function computes the coherency as defined by:

    .. math:: Coh_{ij}(f) = \\frac{S_{ij}(f)}{ \\sqrt{ | S_{ii}(f)S_{jj}(f) | } }

    where :math:`S` is the system spectral matrix.

    """

    coh = np.zeros_like(S)

    for i in range(S.shape[0]):
        for j in range(S.shape[0]):
            coh[i, j, :, :] = S[i, j, :, :] / \
                (np.sqrt(np.abs(S[i, i, :, :] * S[j, j, :, :])))

    return coh


__all__.append('coherency')


def magnitude_squared_coherence(S):
    """Method for computing the Magnitude Squred Coherence.

    Parameters
    ----------
    S : ndarray
        The system spectral matrix in 3D or 4D form, of size
        [nchannels x nchannels x nfreqs x nepochs]

    Returns
    -------
    ndarray
        Magnitude squared coherence matrix, of size [nchannels x nchannels x nfreqs x nepochs]

    Note
    ----

    This function computes this magnitude squared coherence as defined by:

    .. math:: MSC_{ij}(f) = \\frac{ |S_{ij}(f)|^2 }{ \\sqrt{ | S_{ii}(f)S_{jj}(f) | } }

    where :math:`S` is the system spectral matrix. This is closely related to
    the complex valued-coherency.

    """

    return np.power(np.abs(coherency(S)), 2)


__all__.append('magnitude_squared_coherence')


def imaginary_coherence(S):
    """Method for computing the imaginary coherence

    Parameters
    ----------
    S : ndarray
        The system spectral matrix in 3D or 4D form, of size
        [nchannels x nchannels x nfreqs x nepochs]

    Returns
    -------
    ndarray
        Imaginary coherence matrix, of size [nchannels x nchannels x nfreqs x nepochs]

    Note
    ----

    This function computes the imaginary cohereence as defined by:

    .. math:: iCoh_{ij}(f) = \\Im \\bigg( \\frac{ S_{ij}(f) }{ \\sqrt{ | S_{ii}(f)S_{jj}(f) | } }  \\bigg)

    where :math:`S` is the system spectral matrix. This is closely related to
    the complex valued-coherency.


    """

    return coherency(S).imag


__all__.append('imaginary_coherence')


def phase_coherence(S):
    """Method for computing the phase coherence

    Parameters
    ----------
    S : ndarray
        The system spectral matrix in 3D or 4D form, of size
        [nchannels x nchannels x nfreqs x nepochs]

    Returns
    -------
    ndarray
        Phase coherence matrix, of size [nchannels x nchannels x nfreqs x nepochs]

    Note
    ----

    This function computes the phase-coherence as defined by:

    .. math:: pCoh_{ij}(f) = \\arg \\bigg( \\frac{ S_{ij}(f) }{ \\sqrt{ | S_{ii}(f)S_{jj}(f) | } } \\bigg )

    where :math:`S` is the system spectral matrix. This is closely related to
    the complex valued-coherency.


    """

    return np.angle(coherency(S))


__all__.append('phase_coherence')


def partial_coherence(inv_S):
    """Method for computing the Partial Coherence.

    Parameters
    ----------
    inv_S : ndarray
        The system spectral matrix in 3D or 4D form, of size
        [nchannels x nchannels x nfreqs x nepochs]

    Returns
    -------
    ndarray
        Complex-valued coherency matrix, of size [nchannels x nchannels x nfreqs x nepochs]

    Note
    ----

    This function computes the partial coherence using the inverse spectral
    matrix :math:`P = S^{-1}` where :math:`S` is the system spectral matrix.
    The partial coherence is then computed as:

    .. math:: PCoh_{ij}(f) = \\frac{ |P_{ij}(f)|^2 }{ \\sqrt{ | P_{ii}(f)P_{jj}(f) | } }

    """

    pcoh = np.zeros_like(inv_S)

    for i in range(inv_S.shape[0]):
        for j in range(inv_S.shape[0]):
            pcoh[i, j, :, :] = np.power(np.abs(inv_S[i, j, :, :]), 2) / \
                (inv_S[i, i, :, :] * inv_S[j, j, :, :])

    return pcoh


__all__.append('partial_coherence')


def ff_directed_transfer_function(H):
    """Function for computing the full-frequency directed transfer function as defined in [Korzeniewska2003]_.

    Parameters
    ----------
    H : ndarray
        The system transfer function [nchannels x nchannels x nfreqs x nepochs]

    Returns
    -------
    ndarray
        The full-frequency directed transfer function

    Note
    ----

    This function computes the directed transfer function from the system transfer function :math:`H`.

    .. math:: ffDTF_{ij}(f) = \\frac{ | H_{ij}(f) |^2 }{ \\sqrt{ \\sum_f \\sum_f | \\bar{H}_{ij}(f) |^2 } }

    This is similar to the directed transfer function with the exceptiont hat
    the normalisation in the denominator is summed across frequency.

    """

    dtf = np.zeros_like(H)

    for e in range(H.shape[3]):
        for i in range(H.shape[0]):
            # Make the denominator for this channel over all frequencies
            dtf_denom = np.sum(np.power(np.abs(H[i, :, :, e]), 2))

            # Estimate ffDTF
            for f in range(H.shape[2]):
                for j in range(H.shape[0]):
                    dtf[i, j, f, e] = np.power(
                        np.abs(H[i, j, f, e]), 2) / dtf_denom

    return dtf


__all__.append('ff_directed_transfer_function')


def d_directed_transfer_function(H, inv_S):
    """Function for computing the direct directed transfer function as defined in [Korzeniewska2003]_.

    Parameters
    ----------
    H :
        The transfer matrix [nchannels x nchannels x nfreqs x nepochs]
    s :
        inv_S: The inverse spectral matrix
        [nchannels x nchannels x nfreqs x nepochs]
    inv_S :


    Returns
    -------
    ndarray
        The direct directed transfer function

    Note
    ----

    This function computes the direct directed transfer function by point-wise
    multiplicationo of the partial coherence :math:`P` and the full-frequency
    directed transfer function :math:`dDTF`.

    .. math:: dDTF(F) = P(f) * ffDTF(f)

    This is similar to the directed transfer function with the exceptiont hat
    the normalisation in the denominator is summed across frequency.


    """

    return partial_coherence(inv_S) * ff_directed_transfer_function(H)


__all__.append('d_directed_transfer_function')


def partial_directed_coherence(Af):
    """Function to estimate the partial directed coherence from a set of
    multivariate parameters as defined in [Baccala2001]_.

    Parameters
    ----------
    Af : ndarray
        Frequency domain version of parameters (can be calculated using
        ar_spectrum function), of size
        [nchannels x nchannels x nfrequencies x nrealisations]

    Returns
    -------
    ndarray
        The partial directed coherence of the system.

    Note
    ----

    This function computes the partial coherence using the frequency transform
    of the autoregressive parameters :math:`Af` subtracted from Identity.

    .. math:: \\bar{Af}(f) = I - Af(f)

    The partial directed coherence is then

    .. math:: PDC_{ij}(f) = \\frac{ | \\bar{Af}_{ij}(f)| }{ \\sqrt{ \\sum_i | \\bar{Af}_{ij}(f) |^2 } }

    """

    pdc = np.zeros_like(Af)

    # Remove identity from Af
    Afi = np.repeat(np.eye(Af.shape[0])[:, :, None],
                    Af.shape[2], axis=-1)[..., None] - Af

    for e in range(Afi.shape[3]):
        for j in range(Afi.shape[0]):
            for f in range(Afi.shape[2]):
                pdc_denom = np.sqrt(
                    (np.abs(Afi[:, j, f, e])**2).sum())
                for i in range(Afi.shape[0]):
                    pdc[i, j, f, e] = np.abs(
                        Afi[i, j, f, e]) / pdc_denom

    return pdc


__all__.append('partial_directed_coherence')


def isolated_effective_coherence(Af, noise_cov):
    """Function for estimating the Isolated Effective Coherence as defined in
    [Pascual-Marqui2014]_.

    Parameters
    ----------
    Af : ndarray
        Frequency domain version of parameters (can be calculated
        using ar_spectrum function)
        [nchannels x nchannels x nfreqs x nepochs]
    noise_cov : ndarray
        The noise covariance matrix of the system
        [nchannels x nchannels x nepochs]

    Returns
    -------
    ndarray
        The isolated effective coherence.

    """

    # Remove identity from Af
    Afi = np.repeat(np.eye(Af.shape[0])[:, :, None],
                    Af.shape[2], axis=-1)[..., None] - Af

    nchannels = Afi.shape[0]
    nfreqs = Afi.shape[2]
    nepochs = Afi.shape[3]

    iec = np.zeros((nchannels, nchannels, nfreqs, nepochs), dtype=complex)

    for e in range(nepochs):
        # Get inverse of the residual noise covariance and set off-diagonal
        # elements to zero
        S = noise_cov[..., e]  # S = np.linalg.inv(noise_cov[..., e])
        iso_s = np.linalg.inv(np.diag(np.diag(S)))

        for f in range(nfreqs):
            for i in range(nchannels):
                for j in range(nchannels):

                    iso_params = np.diag(np.diag(Afi[:, :, f, e]))
                    # Set all parameters except the diagonal and connection of
                    # interest to zero
                    iso_params[i, j] = Afi[i, j, f, e]

                    denom_col = iso_s[j, j] * np.abs(iso_params[j, j])**2

                    numerator = iso_s[i, i] * np.abs(iso_params[i, j])**2

                    iec[i, j, f, e] = numerator / (numerator + denom_col)

    return iec.real


__all__.append('isolated_effective_coherence')


def directed_transfer_function(H):
    """Method for computing the Directed Transfer Function as defined in
    [Kaminski1991]_.

    Parameters
    ----------
    H : ndarray
        The transfer matrix [nchannels x nchannels x nfreqs x nepochs]

    Returns
    -------
    ndarray
        The directed transfer function of the system.

    Note
    ----

    This function computes the directed transfer function from the system transfer function :math:`H`.

    .. math:: DTF_{ij}(f) = \\frac{ | H_{ij}(f) |^2 }{ \\sqrt{ \\sum_j | \\bar{H}_{ij}(f) |^2 } }

    """

    dtf = np.zeros_like(H)

    for e in range(H.shape[3]):
        for i in range(H.shape[0]):
            for f in range(H.shape[2]):
                # Make the denominator for this channel
                dtf_denom = np.sum(np.power(np.abs(H[i, :, f, e]), 2))
                for j in range(H.shape[0]):
                    dtf[i, j, f, e] = np.power(
                        np.abs(H[i, j, f, e]), 2) / dtf_denom

    return dtf


__all__.append('directed_transfer_function')


def geweke_granger_causality(S, H, sigma):
    """This function computes the Geweke-Granger causality as defined in
    [Barrett2010]_

    Parameters
    ---------
    S : ndarray
        Spectral matrix [nchannels x nchannels x nfreqs x nepochs]
    H : ndarray
        The transfer matrix [nchannels x nchannels x nfreqs x nepochs]
    sigma : ndarray
        Residual noise covariance matrix

    Returns
    -------
    ndarray
        The Geweke-Granger causality


    """

    nchannels = S.shape[0]
    nfreqs = S.shape[2]
    nepochs = S.shape[3]

    ggc = np.zeros((nchannels, nchannels, nfreqs, nepochs), dtype=complex)

    for e in range(nepochs):

        # Compute conditional sigma matrix
        cond_sigma = np.zeros((nchannels, nchannels))
        inv_sigma = np.linalg.inv(sigma[:, :, e])
        for ii in range(nchannels):
            for jj in range(nchannels):
                cond_sigma[ii, jj] = sigma[jj, jj, e] - \
                                sigma[ii, jj, e]*inv_sigma[ii, ii]*sigma[jj, ii, e]

        for f in range(nfreqs):
            for ii in range(nchannels):
                for jj in range(nchannels):

                    denom = H[ii, jj, f, e] * cond_sigma[ii, jj] * H[ii, jj, f, e].conj()
                    denom = np.abs(S[ii, ii, f, e] - denom)

                    ggc[ii, jj, f, e] = np.log(np.abs(S[ii, ii, f, e]) / denom)

    return ggc


__all__.append('geweke_granger_causality')



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import numpy as np

from .stats import find_cov_multitrial

__all__ = []


def approx_model_selection_criteria(A, EF, method='ss'):
    """Estimate Akaike's Information Criterion and Bayes Information Criterion
    from a model fit using an approximation of the loglikelihood.

    Two approximations are available:
        'ss': The sum square of the residuals. This approximation is outlined
        in [Burnham2002]_ and [Chatfield2003]_.
        'det': The determinant of the residual covariance matrix.

    The sum squares method is used as default.

    Parameters
    ----------
    A : ndarray
        Array containing the parameters of a fitted model, of shape
        [channels x channels x maxorder]
    EF : ndarray
        Array containing the residuals of the fitted model, of shape
        [channels x samples x trials]
    method : {'ss','det'}
        string indicating which method should be used to approximate
        the loglikelihood ('ss' or 'det') (Default value = 'ss')

    Returns
    -------
    tuple
        tuple of (aLL, AIC, BIC)
        aLL: Approximation of the loglikelihood (determined by the method
        defined above)
        AIC: Estimate of Akaike's Information Criterion
        BIC: Estimate of the Bayesian Information Criterion

    """

    k = A.reshape(-1).shape[0]  # number of parameters
    N = EF.reshape(-1).shape[0]  # number of observations

    if method == 'ss':
        # Chatfield 2004 - The analysis of time series p 256.
        # Burnham and Anderson 2004 - multimodel inference, understanding AIC
        # and BIC in model selection
        SS = np.power(EF, 2).sum()  # sum of squares
        aLL = N*np.log(SS/N)

    elif method == 'det':
        # Method used in the GCCA toolbox
        # TODO: make this work
        # NOT CURRENTLY WORKING!!!!!
        raise Exception("method == 'det' is not currently working")
        cov = find_cov_multitrial(EF, EF)
        aLL = np.log(np.linalg.det(cov))

    else:
        raise Exception("Unknown method")

    aLL = np.float64(aLL.real)

    # Compute AIC and BIC
    aic = np.float64(aLL + 2*k)

    # No-one else seems to approximate the BIC using the methods above, though
    # we are following the same logic as for the AIC so it's probably fine.
    bic = np.float64(aLL + k*np.log(N))

    return aLL, aic, bic


__all__.append('approx_model_selection_criteria')



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import os
from os.path import abspath, expanduser, isdir, isfile, join, split

import numpy as np

__all__ = []


def generate_pink_roots(power, order=24):
    """Generates the poles of a system with a pink spectrum according to eqn
    116 in [Kasdin1995]_

    Parameters
    ----------
    power : float
        Power of pole must be between -2 and 2.
    order :
        Order of system to generate (Default value = 24)

    Returns
    -------
    ndarray
        Roots of the system (length of order)

    """
    a = np.zeros((order,))
    a[0] = 1
    for k in range(1, order):
        a[k] = (k - 1 - (power/2.)) * (a[k-1] / k)

    return np.roots(a)


__all__.append('generate_pink_roots')


def model_from_roots(roots):
    """Sets up a univariate model based on the roots of a polynomial

    Parameters
    ----------
    roots : ndarray
        Vector of polynomial roots.

    Returns
    -------
    AbstractLinearModel
        Model fit class containing the parameters defined by the input roots

    """

    # We can use the AbstractLinearModel for this as we
    # don't need the fit_model routine
    from sails import AbstractLinearModel

    ret = AbstractLinearModel()
    ret.parameters = -np.poly(roots)[None, None, :]
    ret.resid_cov = np.ones((1, 1, 1))
    ret.delay_vect = np.arange(ret.order + 1)

    return ret


__all__.append('model_from_roots')


def find_support_path():
    """Returns the path to the support files directory"""
    d, f = split(__file__)

    return join(abspath(d), 'support')


__all__.append('find_support_path')


def find_example_path():
    """Returns the path to the tutorial example data directory (if available)"""

    example_path = os.getenv('SAILS_EXAMPLE_DATA', None)

    if example_path is None:
        # Try the user's home directory
        example_path = join(expanduser("~"), "sails-example-data")

    # A sentinel file which we ensure is in the git repo for the example data
    sentinel = join(example_path, 'SAILS_EXAMPLE_DATA')

    # Check for directory and sentinel file
    if not isdir(example_path) or not isfile(sentinel):
        error = """
Error: Could not find the sails example data (looked in
{})

If you wish to use the example data, download it from git
using:

    git clone https://vcs.ynic.york.ac.uk/analysis/sails-example-data

If the git repository is in your home directory, this function
will then find the path.  If you wish to store the data in another
location, set the SAILS_EXAMPLE_DATA environment variable to
the location of the git checkout
""".format(example_path)

        raise Exception(error)

    return example_path


__all__.append('find_example_path')


def set_sail():
    """ """
    import os
    import time
    rows, columns = os.popen('stty size', 'r').read().split()
    cols = int(columns)

    ship = ['    __|__ __|__ ',
            '    )    ))    )\\',
            '    )    ))    )\\\\',
            '    )____))____) \\\\',
            '\\---__|______|____\\\\----',
            '~~;\\_____________/']

    air = ''
    air = [air + ' '*cols for ii in range(6)]
    air = '\n'.join(air) + '\n'

    sea = ''
    sea = [sea + '~'*cols for ii in range(3)]
    sea = '\n'.join(sea)

    msg = list(air+sea)

    for step in range(5, cols-25):
        msg = list(air+sea)
        for ii in range(len(ship)):
            start = step + (cols+1)*(ii+1)
            stop = start + len(ship[ii])
            msg[start:stop] = ship[ii]

        print('\033[2J\033[<3>A')
        print(''.join(msg))
        time.sleep(1.0/15)

    start = cols//2 + cols*(2) - 2
    stop = start+5
    msg[start:stop] = 'SAILS'
    print('\033[2J\033[<3>A')
    print(''.join(msg))
    time.sleep(2)


__all__.append('set_sail')



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colorbar as cb
from mpl_toolkits.axes_grid1 import Grid
from matplotlib import patches

__all__ = []


def root_plot(rts, ax=None, figargs=dict(), plotargs=dict()):
    """Plot a set of roots (complex numbers).

    Parameters
    ----------
    rts : ndarray_like
        Roots to plot
    ax : matplotlib axes handle
        Optional Axes on which to place plot. (Default value = None)
    figargs : dict
        extra arguments to pass to plt.figure (Default value = dict())
    plotargs : dict
        extra arguments to pass to plt.plot (Default value = dict())

    Returns
    -------
    Axes
        Axes object on which plot was drawn

    """

    if 'figsize' not in figargs:
        figargs['figsize'] = (6, 6)

    if ax is None:
        plt.figure(*figargs)
        ax = plt.subplot(111)

    # Unit Circle
    y = np.sin(np.linspace(0, 2*np.pi, 128))
    x = np.cos(np.linspace(0, 2*np.pi, 128))
    ax.plot(x, y, 'k')
    # Inner circles
    ax.plot(.75*x, .75*y, 'k--', linewidth=.2)
    ax.plot(.5*x, .5*y, 'k--', linewidth=.2)
    ax.plot(.25*x, .25*y, 'k--', linewidth=.2)
    ax.grid(True)

    # Arrow annotation
    ax.plot(1.15*x[8:25], 1.15*y[8:25], 'k')
    ax.arrow(1.15*x[23], 1.15*y[23], x[24]-x[23], y[24]-y[23],
             head_width=.05, color='k')

    # Add poles
    ax.plot(rts.real, rts.imag, 'k+', **plotargs)

    # Labels
    ax.set_ylim(-1.2, 1.2)
    ax.set_xlim(-1.2, 1.2)
    ax.set_xlabel('Real')
    ax.set_ylabel('Imaginary')
    ax.annotate('Frequency', xy=(.56, 1.02))
    ax.set_aspect('equal')

    return ax


__all__.append('root_plot')


def plot_diagonal(freq_vect, metric, F=None, title=None, ax=None):
    """Plot the diagonal spectra from a connectivity array. This is generally
    used to summarised the within-channel spectra within a network.

    Parameters
    ----------
    freq_vect : ndarray
        Vector of frequency values indexing the x-axis
    metric : ndarray
        Matrix of connectivity values of size [nchannels x nchannels x nfrequencies]
    F : matplotlib figure handle
        Handle for matplotlib figure to plot in (Default value = None)
    title : str
        String to use as figure title(Default value = None)
    ax : matplotlib axes handle
        Handle for matplotlib axes to plot in (Default value = None)

    """

    if F is None:
        F = plt.figure(figsize=(6, 6))

    if ax is None:
        ax = F.subplots(1)

    for ii in range(metric.shape[0]):
        ax.plot(freq_vect, metric[ii, ii, :, 0])

    if title is not None:
        ax.set_title(title)

    ax.set_xlabel('Frequency')
    ax.grid(True)


def plot_metric_summary(freq_vect, metric, ind=0, F=None, title=None):
    """Plot the within-channel spectra and connectivity matrix for a given
    connectivity metric.

    Parameters
    ----------
    freq_vect : ndarray
        Vector of frequency values indexing the x-axis
    metric : ndarray
        Matrix of connectivity values of size [nchannels x nchannels x nfrequencies]
    ind : int
        Index into frequency dimension to plot connectivity matrix (Default value = 0)
    F : matplotlib figure handle
        Handle for matplotlib figure to plot in (Default value = None)
    title : str
        String to use as figure title(Default value = None)

    """

    F = plt.figure(figsize=(12, 6))
    ax = F.subplots(1, 2)

    plot_diagonal(freq_vect, metric, F=F, ax=ax[0])
    ylimits = ax[0].get_ylim()
    ax[0].vlines(freq_vect[ind], ylimits[0], ylimits[1])

    s = metric[:, :, ind, 0] - np.diag(np.diag(metric[:, :, ind, 0]))
    im = ax[1].imshow(np.abs(s))
    F.colorbar(im)


def plot_vector(metric, x_vect, y_vect=None, x_label=None, y_label=None,
                title=None, labels=None, line_labels=None, F=None,
                cmap=plt.cm.jet, triangle=None, diag=False,
                thresh=None, two_tail_thresh=False, font_size=10,
                use_latex=False):
    """Function for plotting frequency domain connectivity at a single time-point.

    Parameters
    ----------
    metric : ndarray
        matrix containing connectivity values [nsignals x signals x frequencies
        x participants] in which the first dimension refers to source nodes and
        the second dimension refers to target nodes
    x_vect : ndarray
        vector of frequencies to label the x axis
    y_vect : ndarrat
        vector containing the values for the y-axis
    x_label : string [optional]
        label for the x axis (Default value = None)
    y_label : string [optional]
        label for the y axis (Default value = None)
    title : string [optional]
        title for the figure (Default value = None)
    labels : list
        list of node labels for columns and vectors (Default value = None)
    line_labels : list
        list of labels for each separate line (participant
        dimension in metric) (Default value = None)
    F : figurehandle [optional]
        handle of existing figure to plot within (Default value = None)
    triangle : string [optional]
        string to indicate whether only the 'upper' or 'lower'
        triangle of the matrix should be plotted (Default value = None)
    diag : bool [optional]
        flag to indicate whether the diagonal elements should
        be plotted (Default value = False)
    thresh : ndarray [optional]
        matrix containing thresholds to be plotted alongside
        connectivity values [nsignals x nsignals x frequencies] (Default value = None)
    two_tailed_thresh : bool [optional]
        flag to indicate whether both signs (+/-) of the
        threshold should be plotted
    font_size : int [optional]
        override the default font size
    use_latex : bool
        Flag to indicate whether to render text in latex (Default value = False)

    Returns
    -------
    matplotlib figure handle
        Figure handle containing the plot

    """

    # Set up plotting parameters
    matplotlib.rcParams.update({'font.size': font_size})
    if use_latex:
        matplotlib.rcParams['text.latex.preamble'].append(r'\usepackage{amsmath}')
        plt.rc('font', **{'family': 'sans-serif',
                          'sans-serif': ['Helvetica']})
        plt.rc('text', usetex=True)
        plt.rc('font', family='serif')

    if metric.ndim > 3 and metric.shape[3] > 1:
        # We want to plot each line separately
        ppt = metric.shape[3]
    elif metric.ndim > 3 and metric.shape[3] == 1:
        ppt = 1
    elif metric.ndim == 3:
        ppt = 1
        metric = metric[..., None]

    # Sanity check axis labels
    x_vect = x_vect.squeeze()
    if y_vect is not None:
        y_vect = y_vect.squeeze()
    nbSignals = metric.shape[0]

    # If we don't have any labels, make some up
    if labels is None:
        labels = [chr(65 + (x % 26)) for x in range(nbSignals)]

    if line_labels is None:
        line_labels = [chr(97 + (x % 26)) for x in range(ppt)]

    # Make figure if we don't have one
    if F is None:
        F = plt.figure(figsize=(8.3, 5.8))
        plt.axis('off')

    # Get label indices
    x_label_idx = []
    y_label_idx = []
    for g in range(nbSignals * nbSignals):
        i = int(g / nbSignals)
        j = int(g % nbSignals)

        if triangle == 'upper':
            if i == 0:
                x_label_idx.append(g)
            if i == j-1:
                y_label_idx.append(g)

        elif triangle == 'lower':
            if j == 0:
                x_label_idx.append(g)
            if i == j+1:
                y_label_idx.append(g)
        else:
            if i == 0:
                x_label_idx.append(g)
            if j == 0:
                y_label_idx.append(g)

    # Write labels to plot
    ax_pad = 0.25

    # Make grid with appropriate amount of subplots
    grid = Grid(F, 111,
                nrows_ncols=(nbSignals, nbSignals),
                axes_pad=ax_pad,
                share_all=True)

    # Plot up the matrix
    for g in range(nbSignals*nbSignals):
        i = int(g / nbSignals)
        j = int(g % nbSignals)

        if triangle == 'lower' and i < j:
            grid[g].set_visible(False)
            continue
        elif triangle == 'upper' and i > j:
            grid[g].set_visible(False)
            continue

        if diag is False and i == j:
            grid[g].set_visible(False)
            [grid[g].axis[k].set_visible(False) for k in grid[g].axis.keys()]

        if thresh is not None:
            grid[g].plot(x_vect, thresh[i, j, :], 'k--')
            if two_tail_thresh:
                grid[g].plot(x_vect, -thresh[i, j, :], 'k--')

        if ppt > 1 or thresh is not None:
            for p in range(ppt):
                grid[g].plot(x_vect, metric[i, j, :, p],
                             label=line_labels[p])
                grid[g].set_xlim(x_vect[0], x_vect[-1])

        else:
            grid[g].fill_between(x_vect, metric[i, j, :, 0],
                                 0, facecolor='black',
                                 alpha=.9)

        if y_vect is not None:
            grid[g].set_ylim(y_vect[0], y_vect[-1])

        grid[g].grid()

    if line_labels is not None and ppt > 1:
        grid[nbSignals-1].legend(bbox_to_anchor=(1.05, 1), loc=2)

    ymin, ymax = grid[g].get_ylim()

    grid.axes_llc.set_xlim(x_vect[0], x_vect[-1])

    if triangle == 'lower':
        x_x_pos = x_vect[2]
        x_y_pos = ymax*1.2
        y_x_pos = -x_vect[-2]
        y_y_pos = 1.2
        # Plot the labels
        for g in x_label_idx:
            i = int(g / nbSignals)
            j = int(g % nbSignals)

            grid[g].text(y_x_pos, y_y_pos, r'$\mathbf{%s}$' % (labels[i]))

        for g in y_label_idx:
            i = int(g / nbSignals)
            j = int(g % nbSignals)

            grid[g].text(x_x_pos, x_y_pos,
                         r'$\mathbf{%s \rightarrow}$' % (labels[i-1]))
    elif triangle == 'upper':
        x_x_pos = x_vect[:-2].mean()
        x_y_pos = ymax*1.2
        y_x_pos = -x_vect[-2]
        y_y_pos = .5
        # Plot the labels
        for g in x_label_idx:
            i = int(g / nbSignals)
            j = int(g % nbSignals)

            grid[g].text(x_x_pos, x_y_pos,
                         r'$\mathbf{%s \rightarrow}$' % (labels[j]))

        for g in y_label_idx:
            i = int(g / nbSignals)
            j = int(g % nbSignals)

            grid[g].text(y_x_pos, y_y_pos,
                         r'$\mathbf{%s}$' % (labels[i]))
    else:
        x_x_pos = x_vect[2]
        x_y_pos = ymax*1.2
        y_x_pos = -x_vect[:-2].mean() * 1.7
        y_y_pos = ymax * .5
        # Plot the labels
        for g in x_label_idx:
            i = int(g / nbSignals)
            j = int(g % nbSignals)

            grid[g].text(x_x_pos, x_y_pos,
                         r'$\mathbf{%s \rightarrow}$' % (labels[j]))

        for g in y_label_idx:
            i = int(g / nbSignals)
            j = int(g % nbSignals)

            grid[g].text(y_x_pos, y_y_pos,
                         r'$\mathbf{%s}$' % (labels[i]))

    for g in range(nbSignals * nbSignals):
        [k.set_visible(True) for k in grid[g].texts]

    if x_label is not None:
        grid[nbSignals*(nbSignals-1)].set_xlabel(x_label)
    if y_label is not None:
        grid[nbSignals*(nbSignals-1)].set_ylabel(y_label)

    if title is not None:
        plt.suptitle(title)

    return F


def plot_matrix(metric, x_vect, y_vect, x_label=None, y_label=None,
                z_vect=None, title=None, labels=None, F=None, vlines=None,
                cmap=plt.cm.jet, font_size=8, use_latex=False, diag=True):

    """Function for plotting frequency domain connectivity over many time points

    Parameters
    ----------
    metric : ndarray
        matrix containing connectivity values [nsignals x signals x frequencies
        x participants] in which the first dimension refers to source nodes and
        the second dimension refers to target nodes
    x_vect : 1d array
        vector of frequencies to label the x axis
    y_vect : 1d array [optional]
        vector containing the values for the y-axis
    z_vect : 1d array [optional]
        vector containing values for the colour scale (Default value = None)
    x_label : string [optional]
        label for the x axis (Default value = None)
    y_label : string [optional]
        label for the y axis (Default value = None)
    title : string [optional]
        title for the figure (Default value = None)
    labels : list
        list of node labels for columns and vectors (Default value = None)
    F : figurehandle [optional]
        handle of existing figure to plot within (Default value = None)
    vlines : list
        List of x-axis values to plot a dashed vertical line (Default value = None)
    cmap : matplotlib colormap [optional]
        matplotlib.cm.<colormapname> to use for
        colourscale (redundant for plot vector??) (Default value = plt.cm.jet)
    font_size : int [optional]
        override the default font size
    use_latex : bool
        Flag indicating whether to render text in latex (Default value = False)
    diag : bool
        Flag indicating whether to plot the diagonal subplots (Default value = True)

    Returns
    -------
    matplotlib figure handle
        Figure handle containing the plot

    """
    # Set up plotting parameters
    matplotlib.rcParams.update({'font.size': font_size})
    if use_latex:
        matplotlib.rcParams['text.latex.preamble'].append(r'\usepackage{amsmath}')
        plt.rc('font', **{'family': 'sans-serif',
                          'sans-serif': ['Helvetica']})
        plt.rc('text', usetex=True)
        plt.rc('font', family='serif')

    # Sanity check axis labels
    x_vect = x_vect.squeeze()
    y_vect = y_vect.squeeze()
    nbSignals = metric.shape[0]

    # Set up colour scale
    if z_vect is None:

        if diag is True:
            mx = np.abs(metric).max()
        else:
            upper_tri = np.tril(np.abs(metric).max(axis=(2, 3)), k=-1).max()
            lower_tri = np.triu(np.abs(metric).max(axis=(2, 3)), k=1).max()
            mx = np.max([upper_tri, lower_tri])

        if metric.min() < 0:
            cbounds = np.linspace(-mx, mx)
        else:
            cbounds = np.linspace(0, mx)
    else:
        cbounds = z_vect

    # Make figure if we don't have one
    if F is None:
        F = plt.figure(figsize=(8.3, 5.8))
    plt.axis('off')

    # Write labels to plot
    step = 1. / nbSignals

    # Horizontal
    for i in range(len(labels)):
        plt.text((step*i)+.05, 1.01,
                 r'$\mathbf{%s \rightarrow}$' % (labels[i]),
                 fontsize=font_size,
                 rotation='horizontal')

    # Vertical
    for i in range(len(labels), 0, -1):
        plt.text(-.08, (step*i)-step/10 - .06,
                 r'$\mathbf{%s}$' % (labels[-i]),
                 fontsize=font_size,
                 verticalalignment='center',
                 rotation='vertical')

    # Make grid with appropriate amount of subplots
    grid = Grid(F, 111,
                nrows_ncols=(nbSignals, nbSignals),
                axes_pad=0.1,
                share_all=True)

    # Plot up the matrix
    for g in range(nbSignals*nbSignals):
        i = int(g / nbSignals)
        j = int(g % nbSignals)

        if i != j or diag is True:
            grid[g].contourf(x_vect, y_vect, metric[i, j, :, :],
                             np.linspace(cbounds[0], cbounds[-1], 10),
                             cmap=cmap)
            if vlines is not None:
                grid[g].vlines(vlines, y_vect[0], y_vect[-1], linestyles='dashed')
        else:
            grid[g].set_visible(False)

        if j == 0:
            grid[g].set_ylabel(y_label)
        if i == nbSignals-1:
            grid[g].set_xlabel(x_label)

    # Set labels
    grid.axes_llc.set_ylim(y_vect[0], y_vect[-1])

    # Set colourbar
    ax = F.add_axes([.92, .12, .02, .7], visible=True)
    cb.ColorbarBase(ax, boundaries=cbounds, cmap=cmap)

    if title is not None:
        plt.suptitle(title, fontsize=16)

    return F


def plot_netmat(data, colors=None, cnames=None, xtick_pos=None, ytick_pos=None,
                labels=None, ax=None, hregions='top', vregions='right', showgrid=True,
                **kwargs):
    """Function for plotting connectivity matrices

    Any additional keyword arguments will be passed into pcolormesh

    Parameters
    ----------
    data : ndarray
        2D numpy array of netmat data to plot of size [nrois x nrois]
    colors : dict
        dictionary mapping colour names to matplotlib-usable colours. If not
        provided, name from cnames will be used. (Default value = None)
    cnames : list
         list of len(nrois).  List of colour names for each region (Default value = None)
    labels : list
         List of region labels for the X and Y axes (Default value = None)
    ax : matplotlib axis handle
         Matplotlib axes to plot in (Default value = None)

    Additional Parameters
    ---------------------

    xtick_pos : ndarray
         Positions of X ROI division lines in netmat (Default value = None)
    ytick_pos : ndarray
         Positions of Y ROI division lines in netmat (Default value = None)
    hbar_offset : int
        Offset of ROI bar in Y co-ordinates (default: 2)
    vbar_offset : int
        Offset of ROI bar in X co-ordinates (default: 2)
    bar_gap : float
        Gap to put between ROI bars (default: 0.1)
    bar_thickness : float
        Thickness of the ROI bar (default: 2.5)
    label_fontsize : float
        Font size to use for region labels (default: 5)
    hregions : {'top','bottom','off'}
         Location to place horizontal coloured bars indicating regions of ROIs.
         (Default value = 'top')
    vregions : {'left','right','off'}
         Location to place vertical coloured bars indicating regions of ROIs.
         (Default value = 'right')
    showgrid : bool
         Flag to indicate whether to show grid-lines (Default value = True)
    **kwargs :
        Additional arguments are passed to pcolormesh

    Returns
    -------
    matplotlib axes handle
        axes handle containing the plot

    """

    num_rois = data.shape[0]

    hbar_offset = kwargs.pop('hbar_offset', 2)
    vbar_offset = kwargs.pop('vbar_offset', 2)
    bar_gap = kwargs.pop('bar_gap', 0.1)
    bar_thickness = kwargs.pop('bar_thickness', 2.5)
    label_fontsize = kwargs.pop('label_fontsize', 5)

    # Parse the h and vregions parameters
    if hregions not in ['top', 'bottom', 'off', True, False]:
        raise Exception("hregions must be one of top, bottom, off")

    if hregions is True:
        hregions = 'top'
    if hregions is False:
        hregions = 'off'

    if vregions not in ['left', 'right', 'off', True, False]:
        raise Exception("vregions must be one of left, right, off")

    if vregions is True:
        vregions = 'top'
    if vregions is False:
        vregions = 'off'

    # Ensure that we have a set of axes
    if ax is None:
        plt.figure()
        ax = plt.subplot(1, 1, 1)

    # See comment below regarding pcolormesh vs imshow
    ax.pcolormesh(data, **kwargs)
    plt.axis('scaled')

    if cnames is not None:
        start = None

        for k in range(num_rois):
            thiscname = cnames[k]

            if start is None:
                start = k
                curcname = thiscname

            if k != (num_rois - 1):
                nextcname = cnames[k+1]

            if (k == (num_rois - 1)) or (nextcname != curcname):
                # Need to draw and reset start point
                if colors is not None:
                    color = colors[curcname]
                else:
                    color = curcname

                width = bar_thickness

                if vregions == 'right':
                    ll_x = num_rois + vbar_offset
                else:
                    ll_x = -vbar_offset

                if start == 0:
                    ll_y = start
                    height = k - start + 1 - bar_gap
                elif k == (num_rois - 1):
                    ll_y = start + bar_gap
                    height = k - start + 1
                else:
                    ll_y = start + bar_gap
                    height = k - start + 1 - bar_gap - bar_gap

                if vregions != 'off':
                    rect = patches.Rectangle((ll_x, ll_y), width, height,
                                             linewidth=0, facecolor=color,
                                             clip_on=False)
                    ax.add_patch(rect)

                height = bar_thickness

                if hregions == 'top':
                    ll_y = num_rois + hbar_offset
                else:
                    ll_y = -hbar_offset

                if start == 0:
                    ll_x = start
                    width = k - start + 1 - bar_gap
                elif k == (num_rois - 1):
                    ll_x = start + bar_gap
                    width = k - start + 1
                else:
                    ll_x = start + bar_gap
                    width = k - start + 1 - bar_gap - bar_gap

                if hregions != 'off':
                    rect = patches.Rectangle((ll_x, ll_y), width, height,
                                             linewidth=0, facecolor=color,
                                             clip_on=False)
                    ax.add_patch(rect)

                start = None

    if xtick_pos is None:
        ax.set_xticks([])
    else:
        ax.set_xticks(xtick_pos)

    if ytick_pos is None:
        ax.set_yticks([])
    else:
        ax.set_yticks(ytick_pos)

    # Remove the major labels either way
    ax.set_xticklabels([])
    ax.set_yticklabels([])

    if labels is not None:
        ax.set_xticks([x + 0.5 for x in range(num_rois)], minor=True)
        ax.set_yticks([x + 0.5 for x in range(num_rois)], minor=True)
        ax.set_xticklabels(labels, minor=True, fontsize=label_fontsize, rotation=-90)
        ax.set_yticklabels(labels, minor=True, fontsize=label_fontsize)

    plt.xlim(0, num_rois)
    plt.ylim(0, num_rois)

    if showgrid:
        plt.grid(True, color='w', lw=1)

    return ax


__all__.append('plot_netmat')
